"""
Module for fetching and caching ETH balances from web3.

The main class of this module is :class:`BalancesService`.
It is used for fetching ETH balances from web3 and caching them
for subsequent calls.

Example:
    ::

        from web3cat.fetcher.balances import BalancesService

        addresses = [
            "0xd8dA6BF26964aF9D7eEd9e03E53415D37aA96045",
            "0x88e6a0c2ddd26feeb64f039a2c41296fcb3f5640",
        ]
        blocks = [15700000, 15800000]
        service = BalancesService.create()
        response = service.get_balances(addresses, blocks)
        # => going for web3 rpc

        response = service.get_balances(addresses, blocks)
        # => serving from cache

        response = service.get_balance(addresses[0], blocks[0])
        # => serving from cache

"""

from web3cat.fetcher.balances.balance import Balance
from web3cat.fetcher.balances.repo import BalancesRepo
from web3cat.fetcher.balances.service import BalancesService



from __future__ import annotations

from datetime import datetime
from functools import cached_property
from typing import Any, Dict, List

import numpy as np
import polars as pl
from web3 import Web3
from web3.constants import ADDRESS_ZERO
from web3.contract import Contract

from web3cat.data.core import DataCore
from web3cat.fetcher.erc20_metas.erc20_meta import ERC20Meta
from web3cat.fetcher.events import Event


class ERC20Data(DataCore):
    """
    Datasets for ERC20 token.

    Args:
        token: Token name or address
        address_filter: Limit token transfer data only to these addresses.
                        All transfers for mainstream tokens is a big chunk of data.
                        This optimization makes fetches faster. This filter
                        doesn't apply to mints, burns, and total_supply.
        start: Starting timepoint
        end: Ending timepoint
    """

    _start: int | datetime
    _end: int | datetime

    _token: str
    _address_filter: List[str]

    _meta: ERC20Meta | None
    _transfers: pl.DataFrame | None
    _emission: pl.DataFrame | None
    _token_contract: Contract | None

    def __init__(
        self,
        token: str,
        address_filter: List[str] | None,
        start: int | datetime,
        end: int | datetime,
        **kwargs,
    ):
        super().__init__(start, end, **kwargs)  # pylint: disable=too-many-function-args
        self._token = token
        self._address_filter = [addr.lower() for addr in (address_filter or [])]

    @property
    def contract(self) -> Contract:
        """
        A :class:`web3.contract.Contract` for this token
        """
        return self.meta.contract

    @cached_property
    def meta(self) -> ERC20Meta:
        """
        Metadata for tokens (like name, symbol and decimals)
        """
        return self._erc20_metas_service.get(self._token)

    @property
    def address_filter(self) -> List[str]:
        """
        Address filter for this data (transfers only for these addreses)
        """
        return self._address_filter

    @cached_property
    def transfers(self) -> pl.DataFrame:
        """
        Dataframe with transfers for addresses specified by the ``address_filter``.

        **Schema**

        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | Field                | Type                       | Description                                                                  |
        +======================+============================+==============================================================================+
        | ``timestamp``        | :attr:`numpy.int64`        | Timestamp of the transfer event                                              |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``date``             | :class:`datetime.datetime` | Date for the timestamp                                                       |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``block_number``     | :attr:`numpy.int64`        | Block number for this transfer                                               |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``transaction_hash`` | :class:`str`               | Transaction hash for this transfer                                           |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``log_index``        | :attr:`numpy.int64`        | Log index inside the transaction for this transfer                           |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``from``             | :class:`str`               | The address from which erc20 token was sent                                  |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``to``               | :class:`str`               | The address to which erc20 token was sent                                    |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``value``            | :attr:`numpy.float64`      | Transfer value                                                               |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        """
        return self._fetch_transfers(self._address_filter)

    @cached_property
    def emission(self) -> pl.DataFrame:
        """
        All mints and burns.

        **Schema**

        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | Field                | Type                       | Description                                                                  |
        +======================+============================+==============================================================================+
        | ``timestamp``        | :attr:`numpy.int64`        | Timestamp of the transfer event                                              |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``date``             | :class:`datetime.datetime` | Date for the timestamp                                                       |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``block_number``     | :attr:`numpy.int64`        | Block number for this mint / burn                                            |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``transaction_hash`` | :class:`str`               | Transaction hash for this mint / burn                                        |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``log_index``        | :attr:`numpy.int64`        | Log index inside the transaction for this mint / burn                        |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``from``             | :class:`str`               | If non-zero: the burn address                                                |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``to``               | :class:`str`               | If non-zero: the mint address                                                |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``value``            | :attr:`numpy.float64`      | Transfer value                                                               |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        """

        if self.meta.symbol.upper() == "USDT":
            return self._fetch_usdt_emission()
        if self.meta.symbol.upper() == "WETH":
            return self._fetch_weth_emission()
        return self._fetch_transfers([ADDRESS_ZERO])

    @property
    def volume(self) -> pl.DataFrame:
        """
        Dataframe with transfer volumes and balance changes by address.

        **Schema**

        +-------------+-----------------------+----------------------------------------------+
        | Name        | Type                  | Description                                  |
        +=============+=======================+==============================================+
        | ``address`` | :class:`str`          | Address                                      |
        +-------------+-----------------------+----------------------------------------------+
        | ``volume``  | :attr:`numpy.float64` | Volume for the period                        |
        +-------------+-----------------------+----------------------------------------------+
        | ``change``  | :attr:`numpy.float64` | Change for the period (aka net volume)       |
        +-------------+-----------------------+----------------------------------------------+
        """
        df1 = (
            self.transfers[["from", "value"]]
            .groupby("from")
            .agg(pl.col("value").sum())
            .rename({"from": "address", "value": "volume"})
            .with_column((pl.col("volume") * (-1)).alias("change"))
        )
        df2 = (
            self.transfers[["to", "value"]]
            .groupby("to")
            .agg(pl.col("value").sum())
            .rename({"to": "address", "value": "volume"})
            .with_column(pl.col("volume").alias("change"))
        )
        return (
            df1.extend(df2)
            .groupby("address")
            .agg([pl.col("volume").sum(), pl.col("change").sum()])
            .sort(pl.col("volume"), reverse=True)
        )

    def total_supply(
        self,
        timepoints: List[int | datetime],
    ):
        """
        Dataframe with total supply of the token

        **Schema**

        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | Field                | Type                       | Description                                                                  |
        +======================+============================+==============================================================================+
        | ``timestamp``        | :attr:`numpy.int64`        | Timestamp of total supply snapshot                                           |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``date``             | :class:`datetime.datetime` | Date for the timestamp                                                       |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``block_number``     | :attr:`numpy.int64`        | Block number for this total supply snapshot                                  |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``total_supply``     | :attr:`numpy.float64`      | Total supply in natural tokens (e.g. eth for weth, not wei)                  |
        +----------------------+----------------------------+------------------------------------------------------------------------------+

        """
        initial_balance_wei = self._calls_service.get_call(
            self.contract.functions.totalSupply(),
            self.from_block_number - 1,
        ).response
        initial_total_supply = initial_balance_wei / 10**self.meta.decimals
        blocks = sorted(self._resolve_timepoints(timepoints), key=lambda x: x.number)
        block_numbers = [b.number for b in blocks]
        bs = self._accrued_balances(ADDRESS_ZERO, block_numbers, self.emission)
        out = [
            {
                "timestamp": b.timestamp,
                "date": datetime.fromtimestamp(b.timestamp),
                "block_number": b.number,
                "total_supply": initial_total_supply - balance,
            }
            for b, balance in zip(blocks, bs)
        ]
        return pl.DataFrame(
            out,
            {
                "timestamp": pl.UInt64,
                "date": pl.Datetime,
                "block_number": pl.UInt64,
                "total_supply": pl.Float64,
            },
        )

    def balances(
        self,
        addresses: List[str],
        timepoints: List[int | datetime],
    ) -> pl.DataFrame:
        """
        Dataframe with balances for addresses over time.

        Args:
            addresses: The list of addresses
            timepoints: A list of timepoints

        Returns a Dataframe with fields

        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | Field                | Type                       | Description                                                                  |
        +======================+============================+==============================================================================+
        | ``timestamp``        | :attr:`numpy.int64`        | Timestamp for the snapshot of the balance                                    |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``date``             | :class:`datetime.datetime` | Date for the timestamp                                                       |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``block_number``     | :class:`int`               | Number of the block                                                          |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``address``          | :class:`str`               | Ethereum Address                                                             |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``balance``          | :attr:`numpy.float64`      | Balance of an address at the time                                            |
        +----------------------+----------------------------+------------------------------------------------------------------------------+

        """
        addresses = sorted([a.lower() for a in addresses])
        if not self._address_filter is None:
            mismatched = set(addresses).difference(set(self._address_filter))
            if len(mismatched) > 0:
                raise ValueError(
                    f"Addresses {', '.join(mismatched)} are not in address_filter for this set. "
                    "Please add them when initializing ERC20Data."
                )
        initial_balance_calls = [
            self.contract.functions.balanceOf(Web3.to_checksum_address(addr))
            for addr in addresses
        ]
        initial_balances = self._calls_service.get_calls(
            initial_balance_calls, [self.from_block_number]
        )
        blocks = sorted(self._resolve_timepoints(timepoints), key=lambda x: x.number)
        block_numbers = [b.number for b in blocks]
        out = []
        factor = 10**self.meta.decimals
        for i, addr in enumerate(addresses):
            initial_balance = initial_balances[i].response / factor
            bs = self._accrued_balances(addr, block_numbers, self.transfers)
            out += [
                {
                    "timestamp": b.timestamp,
                    "date": datetime.fromtimestamp(b.timestamp),
                    "block_number": b.number,
                    "address": addr,
                    "balance": balance + initial_balance,
                }
                for b, balance in zip(blocks, bs)
            ]
        return pl.DataFrame(
            out,
            {
                "timestamp": pl.UInt64,
                "date": pl.Datetime,
                "block_number": pl.UInt64,
                "address": pl.Utf8,
                "balance": pl.Float64,
            },
        ).sort("block_number")

    def _accrued_balances(
        self, address: str, blocks: List[int], events: pl.DataFrame
    ) -> List[np.float64]:
        if len(blocks) == 0:
            return []

        address = address.lower()
        transfers = (
            events.filter((pl.col("from") == address) | (pl.col("to") == address))
            .filter((pl.col("from") != address) | (pl.col("to") != address))
            .with_column(
                pl.when(pl.col("from") == address)
                .then(pl.col("value") * (-1))
                .otherwise(pl.col("value"))
                .alias("cash_flow")
            )[["block_number", "timestamp", "cash_flow"]]
            .to_dicts()
        )
        j = 0
        balance = 0
        out = []
        for b in blocks:
            while j < len(transfers) and transfers[j]["block_number"] <= b:
                balance += transfers[j]["cash_flow"]
                j += 1
            out.append(balance)
        return out

    def _fetch_usdt_emission(self) -> pl.DataFrame:
        events = []
        owner = self._calls_service.get_call(
            self.contract.functions.owner(), self.from_block_number
        ).response.lower()
        for raw_event in self._events_service.get_events(
            self.contract.events.Issue,
            self.from_block_number,
            self.to_block_number,
        ):
            raw_event.args = {
                "from": ADDRESS_ZERO,
                "to": owner,
                "value": raw_event.args["amount"],
            }
            events.append(raw_event)

        for raw_event in self._events_service.get_events(
            self.contract.events.Redeem,
            self.from_block_number,
            self.to_block_number,
        ):
            raw_event.args = {
                "from": owner,
                "to": ADDRESS_ZERO,
                "value": raw_event.args["amount"],
            }
            events.append(raw_event)
        return self._transform_raw_events(events)

    def _fetch_weth_emission(self) -> pl.DataFrame:
        events = []
        for raw_event in self._events_service.get_events(
            self.contract.events.Deposit,
            self.from_block_number,
            self.to_block_number,
        ):
            raw_event.args = {
                "from": ADDRESS_ZERO,
                "to": raw_event.args["dst"],
                "value": raw_event.args["wad"],
            }
            events.append(raw_event)

        for raw_event in self._events_service.get_events(
            self.contract.events.Withdrawal,
            self.from_block_number,
            self.to_block_number,
        ):
            raw_event.args = {
                "from": raw_event.args["src"],
                "to": ADDRESS_ZERO,
                "value": raw_event.args["wad"],
            }
            events.append(raw_event)
        return self._transform_raw_events(events)

    def _fetch_transfers(self, addresses: List[str]) -> pl.DataFrame:
        events: List[Event] = []

        # Fetch with "from" and "to" filter
        for filters in self._build_argument_filters(addresses):
            fetched_events = self._events_service.get_events(
                self.contract.events.Transfer,
                self.from_block_number,
                self.to_block_number,
                argument_filters=filters,
            )
            events += fetched_events

        return self._transform_raw_events(events)

    def _transform_raw_events(self, events: List[Event]) -> pl.DataFrame:
        block_numbers = list(set(e.block_number for e in events))
        blocks = self._blocks_service.get_blocks(block_numbers)
        ts_index = {b.number: b.timestamp for b in blocks}
        factor = 10**self.meta.decimals
        transfers = pl.DataFrame(
            [self._event_to_row(e, ts_index[e.block_number], factor) for e in events],
            {
                "timestamp": pl.UInt64,
                "date": pl.Datetime,
                "block_number": pl.UInt64,
                "transaction_hash": pl.Utf8,
                "log_index": pl.UInt64,
                "from": pl.Utf8,
                "to": pl.Utf8,
                "value": pl.Float64,
            },
        )
        return transfers.unique(subset=["transaction_hash", "log_index"]).sort(
            pl.col("timestamp")
        )

    def _build_argument_filters(
        self, address_filter: List[str]
    ) -> List[Dict[str, Any] | None, Dict[str, Any] | None]:
        if len(address_filter) == 0:
            return [None]
        return [{"from": address_filter}, {"to": address_filter}]

    def _event_to_row(self, event: Event, ts: int, val_factor: int) -> Dict[str, Any]:
        fr, to, val = list(event.args.values())
        return {
            "timestamp": ts,
            "date": datetime.fromtimestamp(ts),
            "block_number": event.block_number,
            "transaction_hash": event.transaction_hash,
            "log_index": event.log_index,
            "from": fr.lower(),
            "to": to.lower(),
            "value": val / val_factor,
        }



"""
Data for analyzing a portfolio of different tokens.

Examples:
    .. code::

        from datetime import datetime
        from web3cat.data import PortfolioData

        addresses = [
            "0x78605Df79524164911C144801f41e9811B7DB73D",
            "0xBF72Da2Bd84c5170618Fbe5914B0ECA9638d5eb5",
            "0x8EB8a3b98659Cce290402893d0123abb75E3ab28"
        ]
        tokens = ["USDC", "WETH", "ETH", "WBTC"]
        base_tokens = ["USDC", "WETH"]
        portfolio_data = PortfolioData(
            tokens=tokens,
            base_tokens=base_tokens,
            addresses = addresses,
            start = datetime(2022, 6, 1),
            end = datetime(2022, 9, 1),
            numpoints = 4
        )

        # Portfolio value in USDC by address for USDC, WETH, ETH and WBTC holdings
        portfolio_data.breakdown_by_address("USDC")

        # Portfolio value in WETH by address for USDC, WETH, ETH and WBTC holdings
        portfolio_data.breakdown_by_address("WETH")

        # Portfolio value in USDC by token for USDC, WETH, ETH and WBTC holdings
        portfolio_data.breakdown_by_token("USDC")

        # Portfolio value in WETH by token for USDC, WETH, ETH and WBTC holdings
        portfolio_data.breakdown_by_token("WETH")

        # Balance and price data
        portfolio_data.balances_and_prices

        # Balances of a one specific token
        portfolio_data.balances("USDC")
"""

from web3cat.data.portfolios.portfolio import PortfolioData



"""
Base classes for data classes
"""

from datetime import datetime
from functools import cached_property
import time
from typing import List
from web3 import Web3

from web3cat.fetcher.balances import BalancesService
from web3cat.fetcher.blocks import BlocksService
from web3cat.fetcher.blocks.block import Block
from web3cat.fetcher.calls import CallsService
from web3cat.fetcher.erc20_metas import ERC20MetasService
from web3cat.fetcher.events import EventsService


ETH_START_TIMESTAMP = int(time.mktime(datetime(2015, 7, 30).timetuple()))


class DataCore:
    """
    Base class for all data classes
    """

    start: int | datetime
    end: int | datetime
    _balances_service: BalancesService
    _blocks_service: BlocksService
    _calls_service: CallsService
    _erc20_metas_service: ERC20MetasService
    _events_service: EventsService

    def __init__(self, start: int | datetime, end: int | datetime, **kwargs):
        self.start = start
        self.end = end
        service_args = {
            k: kwargs.pop(k)
            for k in [
                "balances_service",
                "blocks_service",
                "calls_service",
                "erc20_metas_service",
                "events_service",
            ]
            if k in kwargs
        }

        if "balances_service" in service_args:
            self._balances_service = service_args["balances_service"]
        else:
            self._balances_service = BalancesService.create(**kwargs)

        if "blocks_service" in service_args:
            self._blocks_service = service_args["blocks_service"]
        else:
            self._blocks_service = BlocksService.create(**kwargs)

        if "calls_service" in service_args:
            self._calls_service = service_args["calls_service"]
        else:
            self._calls_service = CallsService.create(**kwargs)

        if "erc20_metas_service" in service_args:
            self._erc20_metas_service = service_args["erc20_metas_service"]
        else:
            self._erc20_metas_service = ERC20MetasService.create(**kwargs)

        if "events_service" in service_args:
            self._events_service = service_args["events_service"]
        else:
            self._events_service = EventsService.create(**kwargs)

    @cached_property
    def from_block_number(self) -> int:
        """
        Start block number for the data.
        """
        return self._resolve_timepoints([self.start])[0].number

    @cached_property
    def to_block_number(self) -> int:
        """
        End block number for the data.
        """
        return self._resolve_timepoints([self.end])[0].number

    @cached_property
    def from_timestamp(self) -> int:
        """
        Start unix timestamp for the data.
        """

        return self._resolve_timepoints([self.start])[0].timestamp

    @cached_property
    def to_timestamp(self) -> int:
        """
        End unix timestamp for the data.
        """

        return self._resolve_timepoints([self.end])[0].timestamp

    @property
    def from_date(self) -> datetime:
        """
        Start datetime for the data.
        """

        return datetime.fromtimestamp(self.from_timestamp)

    @property
    def to_date(self) -> datetime:
        """
        End datetime for the data.
        """

        return datetime.fromtimestamp(self.to_timestamp)

    @cached_property
    def w3(self) -> Web3:
        """
        An instance of :class:`web3.Web3`
        """
        return self._blocks_service.w3  # pylint: disable=protected-access

    def _resolve_timepoints(self, timepoints: List[int | datetime]) -> List[Block]:
        resolved_dates = []
        for t in timepoints:
            if isinstance(t, datetime):
                t = int(time.mktime(t.timetuple()))
            resolved_dates.append(t)
        timestamps = []
        blocks = []
        for ts in resolved_dates:
            # This works because on each chain block_time > 1s
            # It means that timepoint is a block
            if ts < ETH_START_TIMESTAMP:
                blocks.append(ts)
            else:
                timestamps.append(ts)
        timestamps_idx = {}
        blocks_idx = {}

        resolved = self._blocks_service.get_latest_blocks_by_timestamps(timestamps)
        i = 0
        for ts in timestamps:
            timestamps_idx[ts] = resolved[i]
            i += 1

        resolved = self._blocks_service.get_blocks(blocks)
        i = 0
        for b in blocks:
            blocks_idx[b] = resolved[i]
            i += 1

        return [blocks_idx.get(tp, timestamps_idx.get(tp)) for tp in resolved_dates]



"""
Fetches Gearbox data and returns it as
`polars.Dataframe <https://pola-rs.github.io/polars/py-polars/html/reference/dataframe.html>`_.

Using :class:`GearboxData` you can:

#. Get the list of liquidation events
#. Get dynamic health factor for the position

Examples:

    .. code::

        from datetime import datetime
        from web3cat.data import GearboxData

        data = GearboxData(start = datetime(2022, 1, 1), end = datetime(2022, 12, 12))

        # List all liquidations
        data.liquidations

"""


from web3cat.data.gearbox.gearbox_data import GearboxData



from __future__ import annotations
import json
from typing import Any, Dict, Tuple


class Call:
    """
    Call represents a static call to Ethereum contract function
    """

    #: Ethereum chain_id
    chain_id: int
    #: Contract address for this call
    address: str
    #: Calldata for this call
    calldata: str
    #: Number of the block for this call
    block_number: int
    #: Response received for the call
    response: Dict[str, Any]

    def __init__(
        self,
        chain_id: int,
        address: str,
        calldata: str,
        block_number: int,
        response: Dict[str, Any],
    ):
        self.chain_id = chain_id
        self.address = address.lower()
        self.calldata = calldata.lower()
        self.block_number = block_number
        self.response = response

    @staticmethod
    def from_row(row: Tuple[int, str, str, int, str]) -> Call:
        """
        Deserialize from web3cat.database row

        Args:
            row: database row
        """
        call = Call(*row)
        call.response = json.loads(call.response)
        return call

    def to_row(self) -> Tuple[int, str, str, int, str]:
        """
        Serialize to database row

        Returns:
            database row
        """
        return (
            self.chain_id,
            self.address,
            self.calldata,
            self.block_number,
            json.dumps(self.response),
        )

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert :class:`Call` to dict
        """
        return {
            "chainId": self.chain_id,
            "address": self.address,
            "calldata": self.calldata,
            "blockNumber": self.block_number,
            "response": self.response,
        }

    @staticmethod
    def from_dict(dct: Dict[str, Any]):
        """
        Create :class:`Call` from dict
        """

        return Call(
            chain_id=dct["chainId"],
            address=dct["address"],
            calldata=dct["calldata"],
            block_number=dct["blockNumber"],
            response=dct["response"],
        )

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False

    def __repr__(self):
        return f"Call({json.dumps(self.to_dict())})"



from __future__ import annotations
import json
from typing import List
from web3cat.fetcher.core import Core
from web3cat.fetcher.balances.balance import Balance
from web3cat.fetcher.balances.repo import BalancesRepo
from web3cat.fetcher.utils import json_response, print_progress


class BalancesService(Core):
    """
    Service for getting and caching ETH balances.

    Since there's no easy way of getting ETH balance deltas from
    Web3, this service queries and caches ETH balance
    for every desired date and address.

    Note:
        It's not as effective as querying token balances, based on fetching Transfer events (deltas)
        in batches. Use it with caution.

    **Request/Response flow**

    ::

                +-----------------+                +-------+ +---------------+
                | BalancesService |                | Web3  | | BalancesRepo  |
                +-----------------+                +-------+ +---------------+
        ---------------  |                             |             |
        | Request call |-|                             |             |
        |--------------| |                             |             |
                         |                             |             |
                         | Find Balance                |             |
                         |------------------------------------------>|
                         |                             |             |
                         | If not found: call Web3     |             |
                         |---------------------------->|             |
                         |                             |             |
                         | Save response               |             |
                         |------------------------------------------>|
            -----------  |                             |             |
            | Response |-|                             |             |
            |----------| |                             |             |
                         |                             |             |

    Args:
        balances_repo: An instance of :class:`BalancesRepo`
        kwargs: Args for the :class:`fetcher.core.Core` class
    """

    _balances_repo: BalancesRepo

    def __init__(self, balances_repo: BalancesRepo, **kwargs):
        super().__init__(**kwargs)
        self._balances_repo = balances_repo

    @staticmethod
    def create(**kwargs) -> BalancesService:
        """
        Create an instance of :class:`BalancesService`

        Args:
            kwargs: Args for the :class:`fetcher.core.Core` class

        Returns:
            An instance of :class:`BalancesService`
        """
        balances_repo = BalancesRepo(**kwargs)
        return BalancesService(balances_repo, **kwargs)

    def get_balances(self, addresses: List[str], blocks: List[int]) -> List[Balance]:
        """
        Get ETH balances for a list of blocks and addresses.

        Args:
            addresses: a list of addresses for ETH balances
            blocks: a list of blocks for ETH balances

        Returns:
            A list of :class:`Balance` for addresses and blocks.
            The size of a list = ``len(addresses) * len(blocks)``
        """
        addresses = {addr.lower() for addr in addresses}
        blocks = set(blocks)
        total_number = len(addresses) * len(blocks)
        if total_number == 0:
            return []

        cached_balances = list(
            self._balances_repo.find(addresses, min(blocks), max(blocks) + 1)
        )
        cached_balances_idx = {}
        for bal in cached_balances:
            key = f"{bal.address}|{bal.block_number}"
            cached_balances_idx[key] = bal

        cached_num = 0
        for addr in addresses:
            for block in blocks:
                key = f"{addr}|{block}"
                if key in cached_balances_idx:
                    cached_num += 1

        num_fetch = len(addresses) * len(blocks) - cached_num
        out = []
        i = 0
        for addr in addresses:
            for block in blocks:
                key = f"{addr}|{block}"
                if key in cached_balances_idx:
                    out.append(cached_balances_idx[key])
                    continue
                print_progress(
                    i,
                    num_fetch,
                    f"Fetching {num_fetch} eth balances",
                )
                out.append(self._fetch_balance_and_save(addr, block))
                i += 1

        if num_fetch > 0:
            print_progress(
                num_fetch,
                num_fetch,
                f"Fetching {num_fetch} eth balances",
            )
        return out

    def get_balance(self, address: str, block_number: int) -> Balance:
        """
        Get ETH balance for a block and an address.

        Args:
            address: The address for the ETH balance
            block_number: The block number for which the ETH balance is fetched

        Returns:
            ETH balance
        """
        address = address.lower()
        balances = list(
            self._balances_repo.find([address], block_number, block_number + 1)
        )
        if len(balances) > 0:
            return balances[0]

        return self._fetch_balance_and_save(address, block_number)

    def clear_cache(self):
        """
        Delete all cached ETH balances
        """
        self._balances_repo.purge()
        self._balances_repo.conn.commit()

    def _fetch_balance_and_save(self, address: str, block_number: int) -> Balance:
        resp = json.loads(
            json_response(
                self.w3.eth.get_balance(
                    self.w3.to_checksum_address(address), block_identifier=block_number
                )
            )
        )
        balance_item = Balance(self.chain_id, block_number, address, resp / 10**18)
        self._balances_repo.save([balance_item])
        self._balances_repo.conn.commit()
        return balance_item



from __future__ import annotations
import json
from typing import Any, Dict, Tuple
from web3cat.fetcher.events_indices.index_data import EventsIndexData


class EventsIndex:
    """
    Stores data about fetched events for a specific
    :code:`chain_id`, ``address``, ``event`` and
    :code:`argument_filters`.

    Args:
        chain_id: Ethereum chain_id
        address: Contract address
        event: Event name
        _args: Argument filters (filter for indexed fields of event)
        data: Index data storing blocks for which events were already fetched.
              See :class:`EventsIndexData` for details.
    """

    #: Ethereum chain_id
    chain_id: int
    #: Contract address
    address: str
    #: Event name
    event: str
    _args: Dict[str, Any]
    #: Index data storing blocks for which events were already fetched.
    data: EventsIndexData

    def __init__(
        self,
        chain_id: int,
        address: str,
        event: str,
        args: Dict[str, Any],
        data: EventsIndexData,
    ):
        self.chain_id = chain_id
        self.address = address
        self.event = event
        self.args = args
        self.data = data

    @staticmethod
    def from_row(row: Tuple[int, str, str, str, bytes]) -> EventsIndex:
        """
        Deserialize from web3cat.database row

        Args:
            row: database row
        """
        chain_id, address, event, args_json, raw_data = row
        args = json.loads(args_json)
        data = EventsIndexData.load(raw_data)
        return EventsIndex(chain_id, address, event, args, data)

    def to_row(self) -> Tuple[int, str, str, str, bytes]:
        """
        Serialize to database row

        Returns:
            database row
        """
        return (
            self.chain_id,
            self.address,
            self.event,
            json.dumps(self.args),
            self.data.dump(),
        )

    @property
    def args(self) -> Dict[str, Any] | None:
        """
        Argument filters for the event.

        Arguments filters are sorted by key, and each value (if ``list``)
        is sorted as well. This sorting is required to compare argument filters fast.
        """
        return self._args

    @args.setter
    def args(self, val: Dict[str, Any] | None):
        self._args = EventsIndex.normalize_args(val)

    @staticmethod
    def normalize_args(args: Dict[str, Any] | None) -> Dict[str, Any]:
        """
        Sort argument filters keys and each value (if ``list``).

        Args:
            args: Event argument filters
        """
        if args is None:
            return {}
        out = {}
        for k in sorted(args.keys()):
            v = args[k]
            if isinstance(v, list):
                v = sorted(v)
            out[k] = v
        return out

    def step(self) -> int:
        """
        Returns :const:`contants.BLOCKS_PER_BIT`.
        """
        return self.data.step()

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert :class:`EventsIndex` to dict
        """
        return {
            "chainId": self.chain_id,
            "address": self.address,
            "event": self.event,
            "args": self.args,
            "data": self.data.to_dict(),
        }

    def _dump_args(self) -> str:
        if self.args is None:
            return json.dumps({})
        out = {}
        for k in sorted(self.args.keys()):
            v = self.args[k]
            if isinstance(v, list):
                v = sorted(v)
            out[k] = v
        return json.dumps(out)

    def __repr__(self) -> str:
        return f"EventsIndex({json.dumps(self.to_dict())})"

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False



from __future__ import annotations
from typing import Any, Dict, Tuple
import json


class Balance:
    """
    Balance represents a snapshot of the ETH balance on Ethereum blockchain.
    """

    #: Ethereum chain_id
    chain_id: int
    #: The block number of the balance snapshot
    block_number: int
    #: The address for the ETH balance (always stored lowercase)
    address: str
    #: ETH balance
    balance: int

    def __init__(self, chain_id: int, block_number: int, address: str, balance: int):
        self.chain_id = chain_id
        self.block_number = block_number
        self.address = address.lower()
        self.balance = balance

    @staticmethod
    def from_row(row: Tuple[int, int, str, int]) -> Balance:
        """
        Deserialize from web3cat.database row

        Args:
            row: database row
        """
        return Balance(*row)

    def to_row(self) -> Tuple[int, int, str, int]:
        """
        Serialize to database row

        Returns:
            database row
        """
        return (
            self.chain_id,
            self.block_number,
            self.address,
            self.balance,
        )

    @staticmethod
    def from_dict(dct: Dict[str, Any]):
        """
        Create :class:`Balance` from dctict
        """
        return Balance(
            chain_id=dct["chainId"],
            block_number=dct["blockNumber"],
            address=dct["address"],
            balance=dct["balance"],
        )

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert :class:`Balance` to dict
        """
        return {
            "chainId": self.chain_id,
            "blockNumber": self.block_number,
            "address": self.address,
            "balance": self.balance,
        }

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False

    def __repr__(self):
        return f"Balance({json.dumps(self.to_dict())})"



"""
Module for fetching and caching blocks metadata from web3.

The main class of this module is :class:`BlocksService`.
It is used for retrieving block metadata from web3 and storing it in
cache so that subsequent requests are returned from cache.

Example:
    ::

        from datetime import datetime, timezone
        from web3cat.fetcher.blocks import BlocksService

        service = BlocksService.create()
        latest_block = service.latest_block

        new_year = datetime(2022, 1, 1, tzinfo=timezone.utc)
        new_year_block = service.get_latest_blocks_by_timestamps(int(new_year.timestamp()))[0]
        # => block 13916166

        new_year_block = service.get_latest_blocks_by_timestamps(int(new_year.timestamp()))[0]
        # cached result

        new_year_block = service.get_blocks(new_year_block.number)
        # cached result
"""

from web3cat.fetcher.blocks.block import Block
from web3cat.fetcher.blocks.repo import BlocksRepo
from web3cat.fetcher.blocks.service import BlocksService



"""
Module for fetching ETH balances for addresses.

Examples:

    .. code::

        from datetime import datetime
        from web3cat.data import EtherData

        dates = [datetime(2021, 6, 1), datetime(2021, 7, 1), datetime(2021, 8, 1), datetime(2021, 9, 1)]
        addresses = ["0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2", "0xf977814e90da44bfa03b6295a0616a897441acec"]
        ether_data = EtherData(min(dates), max(dates))

        # Historical Ether balances for addresses
        ether_data.balances(addresses, dates)

"""

from web3cat.data.ethers.ether_data import EtherData



"""
Module for fetching and caching ERC20 tokens metadata
(name, symbol, decimals).

The main class of this module is :class:`ERC20MetasService`.
It fetches token metadata from the preloaded cache or directly
from the blockchain.

Example:
    ::

        from web3cat.fetcher.erc20_metas import ERC20MetasService

        service = ERC20MetasService.create()
        dai_meta = service.get("Dai")
        # => ERC20Meta({"chainId": 1, "address": "0x6b175474e89094c44da98b954eedeac495271d0f", "name": "dai stable coin", "symbol": "dai", "decimals": 18})
        usdc_meta = service.get("USDC")
        # => ERC20Meta({"chainId": 1, "address": "0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48", "name": "usd coin", "symbol": "usdc", "decimals": 6})
        weth_meta = service.get("0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2")
        # => ERC20Meta({"chainId": 1, "address": "0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2", "name": "weth", "symbol": "weth", "decimals": 18})
"""

from web3cat.fetcher.erc20_metas.erc20_meta import ERC20Meta
from web3cat.fetcher.erc20_metas.repo import ERC20MetasRepo
from web3cat.fetcher.erc20_metas.service import ERC20MetasService



"""
Module with repos and models for :class:`EventsIndex`.

:class:`EventsIndex` stores the information about fetched events.
Each :class:`EventsIndex` is persisted in the :code:`events_indices` table.
:class:`EventsIndex` has a field ``data`` that stores the blocks
for which events (with given chain_id, contract address,
event name, and arguments for lookup) were already fetched.

Note:
    :class:`EventsIndex` has some granularity in storing the blocks defined by the
    :const:`constants.BLOCKS_PER_BIT` constant. This means:

    1. You cannot fetch events for less than :const:`constants.BLOCKS_PER_BIT` blocks
    2. Fetch start and end blocks should be snapped to the :const:`constants.BLOCKS_PER_BIT` grid.

Example
~~~~~~~

As an illustratory example, consider how an instance of :class:`EventsIndex`
evolves when fetching the same type of events for different blocks.

Assume :code:`BLOCKS_PER_BIT == 1000`.

**Step1: Fetch events for blocks 11337 - 13798**

First, the fetch blocks are rounded to 11000 - 14000, and events are fetched
for these blocks.
Start timestamp, in this case, should be 11000. However, there's the rule
that start timestamp must be a multiple of 8 * :const:`constants.BLOCKS_PER_BIT`
(for performance reasons). So it's rounded down to 8000.

:class:`EventsIndex`'s ``data`` is updated to ``00001F401C``

+-----------------+------------+-----------------+---------------+
| Timestamp (hex) | Mask (hex) | Timestamp (dec) | Mask (bin)    |
+=================+============+=================+===============+
| 00001F40        | 1C         | 8000            | 00011100      |
+-----------------+------------+-----------------+---------------+

So the first bit of the mask is blocks 8000 - 9000, the second is 9000 - 10000, etc.

**Step2: Fetch blocks 15546 - 19403**

Round blocks to 15000 - 20000, update ``data`` to ``00001F401DF0``

+-----------------+------------+-----------------+------------------+
| Timestamp (hex) | Mask (hex) | Timestamp (dec) | Mask (bin)       |
+=================+============+=================+==================+
| 00001F40        | 1DF0       | 8000            | 0001110111110000 |
+-----------------+------------+-----------------+------------------+

**Step3: Fetch blocks 3000 - 5000**

Now the start timestamp will be updated to 0 (the multiple of 8 * :const:`constants.BLOCKS_PER_BIT`,
rounded down). The new index is ``000000001C1DF0``

+-----------------+------------+-----------------+--------------------------+
| Timestamp (hex) | Mask (hex) | Timestamp (dec) | Mask (bin)               |
+=================+============+=================+==========================+
| 00000000        | 1C1DF0     | 0               | 000111000001110111110000 |
+-----------------+------------+-----------------+--------------------------+

"""

from web3cat.fetcher.events_indices.index import EventsIndex
from web3cat.fetcher.events_indices.index_data import EventsIndexData
from web3cat.fetcher.events_indices.bitarray import BitArray
from web3cat.fetcher.events_indices.repo import EventsIndicesRepo



from __future__ import annotations
import json
from typing import Any, Dict
from web3cat.fetcher.events_indices.constants import BLOCKS_PER_BIT
from web3cat.fetcher.events_indices.bitarray import BitArray


class EventsIndexData:
    """
    Stores blocks numbers with already fetched events.

    All blocks are divided by chunks of
    :const:`constants.BLOCKS_PER_BIT` size. A bit set to ``True``
    means that events for the chunk were already fetched.

    As a storage optimistaion :code:`start_block` parameter is used.
    It is the first block that has ``True`` entries after it.
    Obviously, it must be a multiple of :const:`constants.BLOCKS_PER_BIT`.
    However, there's an additional restriction of :code:`start_block` being
    a multiple of 8 * :const:`constants.BLOCKS_PER_BIT` (for performance
    reasons).

    See the example section in :mod:`fetcher.events_indices` for
    more details.

    Args:
        start_block: The starting block for the index
        mask: A bitset for storing index values
    """

    _start_block: int | None
    end_block: int | None
    _mask: BitArray
    _blocks_per_bit: int

    def __init__(
        self,
        start_block: int | None = None,
        end_block: int | None = None,
        mask: bytes | None = None,
    ):
        self._blocks_per_bit = BLOCKS_PER_BIT
        self._start_block = None
        self.end_block = end_block
        self._mask = BitArray(mask or [])
        self._update_start_block(start_block)

    def set_range(self, start_block: int, end_block: int, value: bool):
        """
        Set blocks range from :code:`start_block` (inclusive)
        to :code:`end_block` (non-inclusive) to ``value``.

        Args:
            start_block: start of the range (inclusive)
            end_block: end of the range (non-inclusive)

        Exceptions:
            The :code:`start_block` and :code:`end_block` arguments
            must be a multiples of :const:`constants.BLOCKS_PER_BIT`
            (:meth:`snap_block_to_grid` can be used for that).
            Otherwise :class:`IndexError` is raised.
        """
        if (
            start_block % self._blocks_per_bit != 0
            or end_block % self._blocks_per_bit != 0
        ):
            raise IndexError(
                f"EventsIndexData only multiples of `{self._blocks_per_bit}` "
                f"are allowed for start_block and end_block. Got `{start_block}` "
                f"and `{end_block}`"
            )
        self._update_start_block(start_block)
        start_idx = self._block_to_bit(start_block)
        end_idx = self._block_to_bit(end_block)
        self._mask.set_range(start_idx, end_idx, value)

    def snap_block_to_grid(self, block: int) -> int:
        """
        Round down block to the nearest chunk start
        (chunks of size :const:`constants.BLOCKS_PER_BIT`)
        """
        return block - block % self._blocks_per_bit

    def step(self) -> int:
        """
        Returns :const:`constants.BLOCKS_PER_BIT`.
        """
        return self._blocks_per_bit

    def dump(self) -> bytes:
        """
        Serialize class data into binary.

        The binary format is

        +-------------+-----------+---------------------+
        | start_block | end_block | mask                |
        +=============+===========+=====================+
        | 4 bytes     | 4 bytes   | n bytes (as needed) |
        +-------------+-----------+---------------------+
        """
        if self._start_block is None:
            return bytes()
        bytes4 = self._start_block.to_bytes(4, "big")
        end_block = self.end_block or 0
        bytes48 = end_block.to_bytes(4, "big")
        return bytes4 + bytes48 + self._mask._data  # pylint: disable=protected-access

    @staticmethod
    def load(data: bytes) -> EventsIndexData:
        """
        Restore class from binary. See :meth:`dump` for binary format.

        Returns:
            An instance of :class:`EventsIndexData`
        """
        if len(data) < 4:
            return EventsIndexData()
        start_block = int.from_bytes(data[0:4], "big")
        end_block = int.from_bytes(data[4:8], "big")
        if end_block == 0:
            end_block = None
        mask = data[8:]
        return EventsIndexData(start_block, end_block, mask)

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert :class:`EventsIndexData` to dict
        """
        return {
            "startBlock": self._start_block,
            "endBlock": self.end_block,
            "mask": self._mask.to_hex(),
        }

    def __getitem__(self, block: int) -> bool:
        """
        Checks if the current block was fetched or not

        Args:
            block: block number

        Returns:
            ``True`` if the block was fetched, ``False`` otherwise
        """
        if self._start_block is None:
            return False
        if not self.end_block is None:
            if block >= self.end_block:
                return False
        bit = self._block_to_bit(block)
        if bit < 0 or bit >= len(self._mask):
            return False
        return self._mask[bit]

    def __repr__(self) -> str:
        return f"EventsIndexData({json.dumps(self.to_dict())})"

    def _update_start_block(self, block: int | None):
        """

        Example

        Before:
        start_block: 0x10
        mask: 0011000011111100
        0               1
        0123456789abcdef0123456789abcdef
                        0011000011111100

        update start block to 0x04
        prepend 2 bytes, update start_block to 0x00 (multiple of 8 bits)

        After:
        start_block: 0x00
        mask: 00000000000000000011000011111100
        0               1
        0123456789abcdef0123456789abcdef
        00000000000000000011000011111100

        """
        if block is None:
            return

        bit_number = block // self._blocks_per_bit
        byte_number = bit_number // 8
        # snap block to grid of bytes and blocks_per_bit
        block = byte_number * 8 * self._blocks_per_bit

        if self._start_block is None:
            self._start_block = block
            return

        if block >= self._start_block:
            # no need to update
            return

        num_bytes_to_prepend = (
            (self._start_block - block) // self._blocks_per_bit
        ) // 8
        self._mask.prepend_empty_bytes(num_bytes_to_prepend)
        self._start_block = block

    def _block_to_bit(self, block: int) -> int:
        return (block - self._start_block) // self._blocks_per_bit

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False



from __future__ import annotations
from functools import cached_property
from datetime import datetime
from typing import Dict, Any, List
import os
import json
import polars as pl
from web3.contract import Contract
from web3.constants import ADDRESS_ZERO

from web3cat.data.core import DataCore

DATA_COMPRESSOR = "0x0a2CA503153Cd5CB2892a0928ac0F71F49a3c194"
ACCOUNT_FACTORY = "0x444cd42baeddeb707eed823f7177b9abcc779c04"
ACCOUNT_FACTORY_DEPLOY_BLOCK = 13810899


class GearboxData(DataCore):
    """
    Data for the gearbox protocol

    Args:
        start: Starting timepoint
        end: Ending timepoint
    """

    def __init__(self, start: int | datetime, end: int | datetime, **kwargs):
        super().__init__(start, end, **kwargs)
        if self._blocks_service.chain_id != 1:
            raise ValueError(
                "For gearbox Mainnet RPC is required (chain_id = 1), got chain_id = {self.blocks_service.chain_id}"
            )

    def credit_account_data(
        self,
        credit_accounts: List[str],
        timepoints: List[int | datetime],
    ) -> pl.DataFrame:
        """
        Health factor for borrower in a credit pool specified by token.

        Args:
            borrower_address: address of the borrower
            token: token that determines the credit pool
            timepoints: a list of timepoints

        Returns:
            A dataframe with fields

        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | Field                | Type                       | Description                                                                  |
        +======================+============================+==============================================================================+
        | ``timestamp``        | :attr:`numpy.int64`        | Timestamp for the snapshot of the balance                                    |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``date``             | :class:`datetime.datetime` | Date for the timestamp                                                       |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``block_number``     | :attr:`numpy.int64`        | Number of the block                                                          |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``token``            | :class:`str`               | Pool token name                                                              |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``facade``           | :class:`str`               | Address of the credit facade                                                 |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``manager``          | :class:`str`               | Address of the credit manager                                                |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``tvl``              | :attr:`numpy.float64`      | Position Total value                                                         |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``health_factor``    | :attr:`numpy.float64`      | Position Health factor                                                       |
        +----------------------+----------------------------+------------------------------------------------------------------------------+

        """
        blocks = self._resolve_timepoints(timepoints)
        blocks_idx = {b.number: b for b in blocks}
        cm_calls = [
            self._credit_account(acc).functions.creditManager()
            for acc in credit_accounts
        ]

        cm_resps = self._calls_service.get_calls(
            cm_calls,
            [self.to_block_number],
        )
        pools = [self._get_pool_by_credit_manager(cm.response) for cm in cm_resps]
        credit_accounts_v1 = set()
        for pool, ca in zip(pools, credit_accounts):
            if pool is None:
                credit_accounts_v1.add(ca)
        credit_accounts = [ca for ca in credit_accounts if not ca in credit_accounts_v1]
        pools = [p for p in pools if not p is None]
        pools_facade_idx = {
            pool["facade"].lower(): {"pool": pool, "credit_account": credit_accounts[i]}
            for i, pool in enumerate(pools)
        }

        out = []
        hf_calls = [
            self._credit_facade(pool["facade"]).functions.calcCreditAccountHealthFactor(
                self.w3.to_checksum_address(acc)
            )
            for pool, acc in zip(pools, credit_accounts)
        ]
        tvl_calls = [
            self._credit_facade(pool["facade"]).functions.calcTotalValue(
                self.w3.to_checksum_address(acc)
            )
            for pool, acc in zip(pools, credit_accounts)
        ]

        hf_resps = self._calls_service.get_calls(
            hf_calls,
            [b.number for b in blocks],
        )
        tvl_resps = self._calls_service.get_calls(
            tvl_calls,
            [b.number for b in blocks],
        )

        out = []
        for hf, tvl in zip(hf_resps, tvl_resps):
            block = blocks_idx[hf.block_number]
            pool = pools_facade_idx[hf.address.lower()]["pool"]
            credit_account = pools_facade_idx[hf.address.lower()]["credit_account"]
            out.append(
                {
                    "timestamp": block.timestamp,
                    "date": datetime.fromtimestamp(block.timestamp),
                    "block_number": block.number,
                    "credit_account": credit_account,
                    "token": pool["token"].symbol.lower(),
                    "facade": pools_facade_idx[hf.address.lower()]["pool"]["facade"],
                    "manager": pools_facade_idx[hf.address.lower()]["pool"]["manager"],
                    "tvl": tvl.response[0] / 10 ** pool["token"].decimals,
                    "health_factor": hf.response / 10**4,
                }
            )
        return pl.DataFrame(
            out,
            {
                "timestamp": pl.UInt64,
                "date": pl.Datetime,
                "block_number": pl.UInt64,
                "credit_account": pl.Utf8,
                "token": pl.Utf8,
                "facade": pl.Utf8,
                "manager": pl.Utf8,
                "tvl": pl.Float64,
                "health_factor": pl.Float64,
            },
        ).sort("block_number")

    @cached_property
    def liquidations(self) -> pl.DataFrame:
        """
        Dataframe with liquidation events.

        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | Field                | Type                       | Description                                                                  |
        +======================+============================+==============================================================================+
        | ``timestamp``        | :attr:`numpy.int64`        | Timestamp for the snapshot of the balance                                    |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``date``             | :class:`datetime.datetime` | Date for the timestamp                                                       |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``block_number``     | :class:`int`               | Number of the block                                                          |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``transaction_hash`` | :class:`str`               | Hash of the liquidation tx                                                   |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``token``            | :class:`str`               | Name of the token for the Gearbox pool                                       |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``borrower``         | :class:`str`               | Address of the liquidated borrower                                           |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``liquidator``       | :class:`str`               | Address of the liquidator                                                    |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``to``               | :class:`str`               | Address of the liquidation funds receiver                                    |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``remaining_funds``  | :attr:`numpy.float64`      | Amount of remaining funds                                                    |
        +----------------------+----------------------------+------------------------------------------------------------------------------+

        """
        events = []
        for pool in self.pools:
            token = pool["token"]
            facade = self._credit_facade(pool["facade"])
            pool_events = self._events_service.get_events(
                facade.events.LiquidateCreditAccount,
                self.from_block_number,
                self.to_block_number,
            )
            for e in pool_events:
                e.args["remainingFunds"] /= 10**token.decimals
                e.args["token"] = token.symbol
            events += pool_events

        blocks = self._resolve_timepoints([e.block_number for e in events])
        out = []
        for b, e in zip(blocks, events):
            out.append(
                {
                    "timestamp": b.timestamp,
                    "date": datetime.fromtimestamp(b.timestamp),
                    "block_number": b.number,
                    "transaction_hash": e.transaction_hash,
                    "token": e.args["token"],
                    "borrower": e.args["borrower"],
                    "liquidator": e.args["liquidator"],
                    "to": e.args["to"],
                    "remaining_funds": e.args["remainingFunds"],
                }
            )
        return pl.DataFrame(
            out,
            {
                "timestamp": pl.UInt64,
                "date": pl.Datetime,
                "block_number": pl.UInt64,
                "transaction_hash": pl.Utf8,
                "token": pl.Utf8,
                "borrower": pl.Utf8,
                "liquidator": pl.Utf8,
                "to": pl.Utf8,
                "remaining_funds": pl.Float64,
            },
        ).sort("block_number")

    def credit_accounts(self, timepoint: int | datetime = None) -> List[str]:
        """
        A list of all credit accounts for a specific timepoint (incl v1)

        Args:
            timepoint: a timepoint to fetch credit accounts for
        """
        if timepoint is None:
            timepoint = self.to_block_number
        to_block = self._resolve_timepoints([timepoint])[0]
        from_block = self._blocks_service.get_blocks([ACCOUNT_FACTORY_DEPLOY_BLOCK])[0]
        initializations = self._events_service.get_events(
            self._account_factory.events.InitializeCreditAccount,
            from_block.number,
            to_block.number,
        )
        returns = self._events_service.get_events(
            self._account_factory.events.ReturnCreditAccount,
            from_block.number,
            to_block.number,
        )

        events = [
            {"kind": "mint", "block": e.block_number, "address": e.args["account"]}
            for e in initializations
        ]
        events += [
            {"kind": "burn", "block": e.block_number, "address": e.args["account"]}
            for e in returns
        ]

        events = sorted(events, key=lambda x: x["block"])
        out = set()
        for e in events:
            if e["kind"] == "mint":
                out.add(e["address"].lower())
            else:
                out.remove(e["address"].lower())
        return sorted(list(out))

    def _get_pool_by_credit_manager(self, credit_manager: str) -> Dict[str, Any] | None:
        for pool in self.pools:
            if pool["manager"].lower() == credit_manager.lower():
                return pool
        return None

    def _credit_account(self, address: str) -> Contract:
        return self.w3.eth.contract(
            address=self.w3.to_checksum_address(address),
            abi=self._credit_account_abi,
        )

    def _credit_facade(self, address: str) -> Contract:
        return self.w3.eth.contract(
            address=self.w3.to_checksum_address(address),
            abi=self._credit_facade_abi,
        )

    def _credit_manager(self, address: str) -> Contract:
        return self.w3.eth.contract(
            address=self.w3.to_checksum_address(address),
            abi=self._credit_manager_abi,
        )

    @cached_property
    def pools(self) -> List[Dict[str, Any]]:
        """
        Gearbox pools data.
        """
        credit_manager_data = self._calls_service.get_call(
            self._data_compressor.functions.getCreditManagersList(),
            self.to_block_number,
        )
        out = []
        for entry in credit_manager_data.response:
            if entry[14] == ADDRESS_ZERO:
                # skip v1
                continue
            out.append(
                {
                    "token": self._erc20_metas_service.get(entry[1]),
                    "manager": entry[0],
                    "facade": entry[14],
                    "configurator": entry[15],
                    "pool": entry[2],
                    "borrow_rate": entry[5] / 10**27,
                    "fee_interest": entry[21] / 10**4,
                    "fee_liquidation": entry[22] / 10**4,
                    "liquidation_discount": entry[23] / 10**4,
                    "fee_liquidation_expired": entry[24] / 10**4,
                    "liquidation_discount_expired": entry[25] / 10**4,
                }
            )
        return out

    @cached_property
    def _account_factory(self) -> Contract:
        current_folder = os.path.realpath(os.path.dirname(__file__))
        abi = None
        with open(
            f"{current_folder}/abi/account_factory_abi.json", "r", encoding="utf-8"
        ) as f:
            abi = json.load(f)
        return self.w3.eth.contract(
            address=self.w3.to_checksum_address(ACCOUNT_FACTORY),
            abi=abi,
        )

    @cached_property
    def _data_compressor(self) -> Contract:
        current_folder = os.path.realpath(os.path.dirname(__file__))
        abi = None
        with open(
            f"{current_folder}/abi/data_compressor_abi.json", "r", encoding="utf-8"
        ) as f:
            abi = json.load(f)
        return self.w3.eth.contract(
            address=self.w3.to_checksum_address(DATA_COMPRESSOR),
            abi=abi,
        )

    @cached_property
    def _credit_account_abi(self) -> Dict[str, Any] | None:
        current_folder = os.path.realpath(os.path.dirname(__file__))
        with open(
            f"{current_folder}/abi/credit_account_abi.json", "r", encoding="utf-8"
        ) as f:
            return json.load(f)

    @cached_property
    def _credit_facade_abi(self) -> Dict[str, Any] | None:
        current_folder = os.path.realpath(os.path.dirname(__file__))
        with open(
            f"{current_folder}/abi/credit_facade_abi.json", "r", encoding="utf-8"
        ) as f:
            return json.load(f)

    @cached_property
    def _credit_manager_abi(self) -> Dict[str, Any] | None:
        current_folder = os.path.realpath(os.path.dirname(__file__))
        with open(
            f"{current_folder}/abi/credit_manager_abi.json", "r", encoding="utf-8"
        ) as f:
            return json.load(f)



"""
Visualization for blockchain data.

The workflow for the visualization:

    1. Fetch the data using classes from :mod:`data`
    2. Visualize the data using `bokeh <https://bokeh.org>`_ lib

All data is cached and can be accessed from the :class:`View`

Examples:

    **Preliminary setup**

    .. code::

        from datetime import datetime
        from bokeh.io import output_notebook
        from web3cat.view import View

        # for python notebooks
        output_notebook()

    **Example 1**

    .. code::

        v1 = View(token="DAI", start=datetime(2022, 6, 1), end = datetime(2022, 10, 30)) \
            .total_supply() \
            .balance(["0x5d3a536E4D6DbD6114cc1Ead35777bAB948E3643", "0x467194771dAe2967Aef3ECbEDD3Bf9a310C76C65"])
        v1.show()

    .. image:: images/view1.png

    **Example 2**

    .. code::

        v2 = View(start=datetime(2022, 6, 1), end = datetime(2022, 10, 30)) \
            .total_supply(token="USDC") \
            .total_supply(token="DAI")
        v2.show()

    .. image:: images/view2.png

    **Example 3**

    .. code::

        v3 = View(token="WBTC", start=datetime(2022, 6, 1), end = datetime(2022, 10, 30)) \
            .total_supply() \
            .chainlink_prices(base_token = "USDC")
        v3.show()

    .. image:: images/view3.png

    **Example 4**

    .. code::

        v4 = View(token="ETH", start=datetime(2022, 6, 1), end = datetime(2022, 10, 30)) \
            .balance(["0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2", "0xf977814e90da44bfa03b6295a0616a897441acec"])
        v4.show()

    .. image:: images/view4.png

    **Example 5**

    .. code::

        addresses = [
            "0x78605Df79524164911C144801f41e9811B7DB73D",
            "0xBF72Da2Bd84c5170618Fbe5914B0ECA9638d5eb5",
            "0x8EB8a3b98659Cce290402893d0123abb75E3ab28"
        ]
        tokens = ["USDC", "WETH", "ETH", "WBTC"]
        base_tokens = ["USDC", "WETH"]

        v5 = View().portfolio_by_token(
            addresses = addresses,
            tokens = tokens,
            base_token = "USDC",
            start=datetime(2022, 10, 1),
            end=datetime(2022, 10, 30),
            numpoints=100
        )
        v5.show()

        v6 = View().portfolio_by_address(
            addresses = addresses,
            tokens = tokens,
            base_token = "USDC",
            start=datetime(2022, 10, 1),
            end=datetime(2022, 10, 30),
            numpoints=100
        )
        v6.show()

        v7 = View().portfolio_balance(
            addresses = addresses,
            base_token = "USDC",
            start=datetime(2022, 10, 1),
            end=datetime(2022, 10, 30),
            numpoints=100
        )
        v7.show()

    .. image:: images/view5.png
    .. image:: images/view6.png
    .. image:: images/view8.png

    **Example 6**

    .. code::

        v6.get_data(0).breakdown_by_token("USDC")

    .. image:: images/view7.png
"""

from web3cat.view.view import View



from typing import Iterator, List
from web3cat.fetcher.balances.balance import Balance
from web3cat.fetcher.core import Core


class BalancesRepo(Core):
    """
    Reading and writing :class:`Balance` to database cache.
    """

    def find(
        self,
        addresses: List[str],
        from_block: int = 0,
        to_block: int = 2**32 - 1,
    ) -> Iterator[Balance]:
        """
        Find all balances in the database cache.

        Args:
            addresses: Contract / EOA addresses
            from_block: starting from this block (inclusive)
            to_block: ending with this block (non-inclusive)

        Returns:
            An iterator over found balances
        """
        if len(addresses) == 0:
            return iter([])

        addresses = [addr.lower() for addr in addresses]
        rows = self.conn.execute(
            f"SELECT * FROM balances WHERE chain_id = ? \
                AND address IN ({','.join(['?'] * len(addresses))}) \
                AND block_number >= ? AND block_number < ?",
            [self.chain_id, *addresses, from_block, to_block],
        )
        return (Balance.from_row(r) for r in rows)

    def save(self, balances: List[Balance]):
        """
        Save a list of balances into the database cache.

        Args:
            balances: List of balances to save
        """
        rows = [e.to_row() for e in balances]
        self.conn.executemany(
            "INSERT INTO balances VALUES(?,?,?,?) ON CONFLICT DO NOTHING", rows
        )

    def purge(self):
        """
        Clear all balances entries from the database cache.
        """
        self.conn.execute("DELETE FROM balances")



from __future__ import annotations
import json
from typing import Any, Dict, List, Tuple
from web3.contract import ContractEvent
from requests.exceptions import ReadTimeout

from web3cat.fetcher.core import Core
from web3cat.fetcher.blocks import BlocksService
from web3cat.fetcher.events.event import Event
from web3cat.fetcher.events.repo import EventsRepo
from web3cat.fetcher.events_indices.index import EventsIndex
from web3cat.fetcher.events_indices.index_data import EventsIndexData
from web3cat.fetcher.events_indices.repo import EventsIndicesRepo
from web3cat.fetcher.utils import json_response, print_progress, short_address


class EventsService(Core):
    """
    Service for fetching contract events.

    This service fetches contract events from web3, cache them,
    and read from the cache on subsequent calls.

    **Request/Response flow**

    ::

                   +---------------+              +-------+ +-------------------+ +-------------+
                   | EventsService |              | Web3  | | EventsIndicesRepo | | EventsRepo  |
                   +---------------+              +-------+ +-------------------+ +-------------+
        -----------------  |                          |               |                  |
        | Request events |-|                          |               |                  |
        |----------------| |                          |               |                  |
                           |                          |               |                  |
                           | Request index            |               |                  |
                           |----------------------------------------->|                  |
                           |                          |               |                  |
                           | Fetch missing events     |               |                  |
                           |------------------------->|               |                  |
                           |                          |               |                  |
                           | Store missing events     |               |                  |
                           |------------------------------------------------------------>|
                           |                          |               |                  |
                           | Fetch all events         |               |                  |
                           |------------------------------------------------------------>|
          ---------------  |                          |               |                  |
          | Response     |-|                          |               |                  |
          |--------------| |                          |               |                  |
                           |                          |               |                  |

    Args:
        events_repo: Repo of events
        events_indices_repo: Repo of events_indices
        kwargs: Args for the :class:`fetcher.core.Core`
    """

    _events_repo: EventsRepo
    _events_indices_repo: EventsIndicesRepo
    _blocks_service: BlocksService

    def __init__(
        self,
        events_repo: EventsRepo,
        events_indices_repo: EventsIndicesRepo,
        blocks_service: BlocksService,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._events_repo = events_repo
        self._events_indices_repo = events_indices_repo
        self._blocks_service = blocks_service

    @staticmethod
    def create(**kwargs) -> EventsService:
        """
        Create an instance of :class:`EventsService`

        Args:
            kwargs: Args for the :class:`fetcher.core.Core`

        Returns:
            An instance of :class:`EventsService`
        """
        events_repo = EventsRepo(**kwargs)
        events_indices_repo = EventsIndicesRepo(**kwargs)
        blocks_service = BlocksService.create(**kwargs)
        return EventsService(
            events_repo, events_indices_repo, blocks_service=blocks_service
        )

    def get_events(
        self,
        event: ContractEvent,
        from_block: int,
        to_block: int,
        argument_filters: Dict[str, Any] | None = None,
    ) -> List[Event]:
        """
        Get events specified by parameters.

        Args:
            event: class:`web3.contract.ContractEvent` specifying contract and event_name.
            from_block: fetch events from this block (inclusive)
            to_block: fetch events from this block (non-inclusive)
            argument_filters: Additional filters for events search.
                              Example: :code:`{"from": "0xfa45..."}`

        Returns:
            A list of fetched events

        Exceptions:
            See :meth:`prefetch_events`
        """
        self.prefetch_events(event, from_block, to_block, argument_filters)
        all_events = self._events_repo.find(
            event.event_name,
            event.address.lower(),
            from_block=from_block,
            to_block=to_block,
            argument_filters=argument_filters,
        )
        return list(all_events)

    def prefetch_events(
        self,
        event: ContractEvent,
        from_block: int,
        to_block: int,
        argument_filters: Dict[str, Any] | None = None,
    ):
        """
        Fetch events specified by parameters and save them to cache.

        Args:
            event: class:`web3.contract.ContractEvent` specifying contract and event_name.
            from_block: fetch events from this block (inclusive)
            to_block: fetch events from this block (non-inclusive)
            argument_filters: Additional filters for events search.
                              Example: :code:`{"from": "0xfa45..."}`

        Exceptions:
            This method tries to fetch all events from :code:`from_block`
            to :code:`to_block` at once. More often than not, rpc endpoint
            will block such attempts and ask to use a narrower block interval
            for fetch. In this case, the interval is halved, and fetch is retried.
            This is repeated until success.

            However, if at some point the interval is less than
            :const:`fetcher.events_indices.constants.BLOCKS_PER_BIT`,
            :class:`RuntimeError` is raised.
        """
        read_indices = list(
            self._events_indices_repo.find_indices(
                event.address, event.event_name, argument_filters
            )
        )
        write_index = self._events_indices_repo.get_index(
            event.address, event.event_name, argument_filters
        )
        if write_index is None:
            write_index = EventsIndex(
                self.chain_id,
                event.address,
                event.event_name,
                argument_filters,
                EventsIndexData(),
            )
        current_chunk_size_in_steps = (to_block - from_block) // write_index.step() + 1
        fetched = False
        while not fetched:
            if current_chunk_size_in_steps == 0:
                raise RuntimeError(
                    "Couldn't fetch data because minimum chunk size is reached"
                )
            try:
                self._fetch_events_for_chunk_size(
                    current_chunk_size_in_steps,
                    event,
                    argument_filters,
                    from_block,
                    to_block,
                    read_indices,
                    write_index,
                )
                fetched = True
            except (ValueError, ReadTimeout):
                current_chunk_size_in_steps //= 2

    def clear_cache(self):
        """
        Delete all cached entries
        """
        self._events_indices_repo.purge()
        self._events_repo.purge()
        self._events_indices_repo.conn.commit()
        self._events_repo.conn.commit()

    def _fetch_events_for_chunk_size(
        self,
        chunk_size_in_steps: int,
        event: ContractEvent,
        argument_filters: Dict[str, Any] | None,
        from_block: int,
        to_block: int,
        read_indices: List[EventsIndex],
        write_index: EventsIndex,
    ):
        from_block = write_index.data.snap_block_to_grid(from_block)
        offset = 0 if to_block == write_index.data.snap_block_to_grid(to_block) else 1
        to_block = (
            write_index.data.snap_block_to_grid(to_block) + offset * write_index.step()
        )
        prefix = (
            f"Fetching {event.event_name}@{short_address(event.address)} "
            f"({from_block} - {to_block})"
        )
        step = chunk_size_in_steps * write_index.step()
        to_fetch = []
        for start in range(from_block, to_block, step):
            end = min(start + step, to_block)
            shrinked_start, shrinked_end = self._shrink_blocks(read_indices, start, end)
            if shrinked_start < shrinked_end:
                to_fetch.append([shrinked_start, shrinked_end])

        if len(to_fetch) == 0:
            return

        for i, tup in enumerate(to_fetch):
            shrinked_start, shrinked_end = tup
            print_progress(
                i,
                len(to_fetch),
                prefix=prefix,
            )
            self._fetch_and_save_events_in_one_chunk(
                event,
                argument_filters,
                shrinked_start,
                shrinked_end,
                write_index,
            )

        print_progress(
            len(to_fetch),
            len(to_fetch),
            prefix=prefix,
        )

    def _shrink_blocks(
        self, read_indices: List[EventsIndex], from_block: int, to_block: int
    ) -> Tuple[int, int]:
        """
        Shorten a block range so that both start and end are facing 0 bits.

        Example:
          from            to            from  to
            |             |               |   |
            111111100101111   =>   111111100101111
        """
        if len(read_indices) == 0:
            return (from_block, to_block)
        step = read_indices[0].step()
        shrinked_start, shrinked_end = from_block, to_block
        while (
            self._block_is_in_indices(read_indices, shrinked_start)
            # this one is for the case of latest block
            and self._block_is_in_indices(read_indices, shrinked_start + step - 1)
            and shrinked_start <= shrinked_end
        ):
            shrinked_start += step
        while (
            self._block_is_in_indices(read_indices, shrinked_end - step)
            and shrinked_start <= shrinked_end
        ):
            shrinked_end -= step
        return (shrinked_start, shrinked_end)

    def _block_is_in_indices(self, indices: List[EventsIndex], block: int) -> bool:
        if block < 0:
            return False
        for i in indices:
            if i.data[block]:
                return True
        return False

    def _fetch_and_save_events_in_one_chunk(
        self,
        event: ContractEvent,
        argument_filters: Dict[str, Any] | None,
        from_block: int,
        to_block: int,
        write_index: EventsIndex,
    ) -> List[Event]:
        latest_number = self._blocks_service.latest_block.number
        if from_block > latest_number:
            return []

        latest_to_grid_block = to_block
        # we're fetching beyond latest block

        if to_block > latest_number + 1:
            to_block = latest_number + 1
            latest_to_grid_block = write_index.data.snap_block_to_grid(to_block)
            if latest_to_grid_block != to_block:
                latest_to_grid_block += write_index.data.step()
            # the latest block in the index is outdated

            if latest_number + 1 > (write_index.data.end_block or 0):
                write_index.data.end_block = latest_number + 1
        events = self._fetch_events_in_one_chunk(
            event, from_block, to_block, argument_filters
        )
        self._events_repo.save(events)
        write_index.data.set_range(from_block, latest_to_grid_block, True)
        self._events_indices_repo.save([write_index])
        self._events_indices_repo.conn.commit()
        return events

    def _fetch_events_in_one_chunk(
        self,
        event: ContractEvent,
        from_block: int,
        to_block: int,
        argument_filters: Dict[str, Any] | None,
    ) -> List[Event]:
        event_filter = event.create_filter(
            fromBlock=from_block,
            toBlock=to_block - 1,
            argument_filters=argument_filters,
        )
        entries = event_filter.get_all_entries()
        jsons = [json.loads(json_response(e)) for e in entries]
        events = [
            Event(
                chain_id=self.chain_id,
                block_number=j["blockNumber"],
                transaction_hash=j["transactionHash"],
                log_index=j["logIndex"],
                address=j["address"],
                event=j["event"],
                args=j["args"],
            )
            for j in jsons
        ]
        return events



"""
Implements :class:`Core` that is used in other modules.
"""

import os
from os.path import exists
from sqlite3 import Connection, connect
from functools import cached_property
from web3 import Web3

DEFAULT_BLOCK_GRID_STEP = 1000
web3_cache = {}
db_cache = {}
chain_id_cache = {}


class Core:
    """
    A base class for any class that wants to use
    an Ethereum RPC or Sqlite3 cache database.

    When deriving this class, you're providing arguments like rpc url
    or OS path to the database. The resources are instantiated
    on demand though. It means that if you're just using the Ethereum
    RPC it's sufficient to supply only the rpc endpoint and skip OS path
    to the database in the constructor.

    So this class lightweight and safe to derive from any other
    class.

    **Caching**

    The web3 instance and chain_id are cached by the rpc url key.
    The sqlite3 connection is cached by the OS path of the database.

    While this might not work well in a multi-threaded scenario, for
    single-threaded there's no overhead like making new connections
    and, for example, querying chain_id each time it's accessed.

    **Block grid**

    It's often desirable to convert block number to timestamp and vice
    versa. In a way, blocks are blockchain-readable, and timestamps are
    human-readable.

    However, fetching every single block is impractical in many cases.

    That's why the following algorithm is used for timestamp estimation:

        1. We make a block number grid with a width specified by the ``block_grid_step`` parameter.
        2. For each block number, we take the two closest grid blocks (below and above).
        3. Fetch the grid blocks
        4. Assume :math:`a_n` and :math:`a_t` is a number
           and a timestamp for the block above
        5. Assume :math:`b_n` and :math:`b_t` is a number
           and a timestamp for the block below
        6. Assume :math:`c_n` and :math:`c_t` is a number
           and a timestamp for the block we're looking for
        7. :math:`w = (c_n - b_n) / (a_n - b_n)`
        8. Then :math:`c_t = b_t \\cdot (1-w) + a_t * w`

    This algorithm gives a reasonably good approximation for the block
    timestamp and considerably reduces the number of block fetches.
    For example, if we have 500 events happening in the 1000 - 2000
    block range, then we fetch only two blocks (1000, 2000) instead of 500.

    If you still want the exact precision, use
    ``block_grid_step = 1``.

    Warning:
        It's highly advisable to use a single ``block_grid_step`` for all data.
        Otherwise (in theory) the happens-before relationship might
        be violated for the data points.

    Args:
        rpc: An https Ethereum RPC endpoint uri
        cache_path: OS path to the cache database
        block_grid_step: Distance between two adjacent grid blocks
        w3: an instance of web3 (overrides rpc)
        conn: an instance of database connection (overrides cache_path)
    """

    #: An https Ethereum RPC endpoint uri.
    _rpc: str | None
    #: OS path to the cache database.
    #: Can be ``None`` if :class:`sqlite3.Connection` is injected directly.
    cache: str | None
    _block_grid_step: int

    def __init__(
        self,
        rpc: str | None = None,
        cache_path: str | None = None,
        block_grid_step: int = DEFAULT_BLOCK_GRID_STEP,
        w3: Web3 | None = None,
        conn: Connection | None = None,
    ):
        self._rpc = rpc
        self.cache_path = cache_path
        self._block_grid_step = block_grid_step
        self._w3 = w3
        self._conn = conn

    @cached_property
    def block_grid_step(self) -> int:
        """
        Distance between two adjacent grid blocks
        """
        env_value = os.environ.get("WEB3_BLOCK_GRID_STEP")
        if not env_value is None:
            return env_value
        return self._block_grid_step

    @cached_property
    def chain_id(self):
        """
        Chain id for the current web3 connection
        """
        if not self.rpc:
            return self.w3.eth.chain_id

        if not self.rpc in chain_id_cache:
            chain_id_cache[self.rpc] = self.w3.eth.chain_id

        return chain_id_cache[self.rpc]

    @cached_property
    def rpc(self) -> str:
        """
        Ethereum RPC endpoint.

        Note:
            Can be ``None`` if :class:`web3.Web3` is injected directly.
        """

        return self._rpc or os.environ.get("WEB3_PROVIDER_URI")

    @cached_property
    def w3(self) -> Web3:
        """
        :class:`web3.Web3` instance for working with Ethereum RPC
        """
        if not self._w3 is None:
            return self._w3

        if self.rpc is None:
            raise ValueError(
                "Ethereum RPC is not set. "
                "Use `WEB3_PROVIDER_URI` env variable or pass rpc explicitly"
            )

        if not self.rpc in web3_cache:
            provider = None
            if self.rpc.lower().startswith("http"):
                provider = Web3.HTTPProvider(self.rpc)
            else:
                provider = Web3.WebsocketProvider(self.rpc, websocket_timeout=60)
            web3_cache[self.rpc] = Web3(provider)

        return web3_cache[self.rpc]

    @cached_property
    def conn(self) -> Connection:
        """
        :class:`sqlite3.Connection` to a database cache
        """
        if not self._conn is None:
            return self._conn

        if self.cache_path is None:
            self.cache_path = os.environ.get("WEB3_CACHE_PATH")

        if self.cache_path is None:
            raise ValueError(
                "Cache database path is not set. "
                "Use `WEB3_CACHE_PATH` env variable or pass cache_path explicitly"
            )

        if not self.cache_path in db_cache:
            db_cache[self.cache_path] = connection_from_path(self.cache_path)

        return db_cache[self.cache_path]


def connection_from_path(path: str) -> Connection:
    """
    Creates a connection to a database at ``path``.
    If the file at ``path`` doesn't exist, creates a new one and
    initializes a database schema.

    Args:
        path: The absolute path to the database

    Returns:
        An instance of sqlite3 Connection

    Note:
        The schema migrations are currently not supported.
    """

    is_fresh = not exists(path)
    conn = connect(path)
    if is_fresh:
        _init_db(conn)

    return conn


def _init_db(conn: Connection):
    """
    Initialize db schema

    Args:
        conn: Connection to the database
    """
    cursor = conn.cursor()
    # Events table
    cursor.execute(
        """CREATE TABLE IF NOT EXISTS events
            (chain_id integer, block_number integer, transaction_hash text,
            log_index integer, address text, event text, args text)"""
    )
    cursor.execute(
        """CREATE UNIQUE INDEX IF NOT EXISTS idx_events_id \
        ON events(chain_id,transaction_hash,log_index)
    """
    )
    cursor.execute(
        """CREATE INDEX IF NOT EXISTS idx_events_search \
        ON events(chain_id,block_number,event,address)
    """
    )

    # Blocks table
    cursor.execute(
        """CREATE TABLE IF NOT EXISTS blocks
                (chain_id integer, block_number integer, timestamp integer)"""
    )
    cursor.execute(
        """CREATE UNIQUE INDEX IF NOT EXISTS idx_blocks_id
            ON blocks(chain_id,block_number)"""
    )

    # Event indices table

    cursor.execute(
        """CREATE TABLE IF NOT EXISTS events_indices
                (chain_id integer, address text, event text, args text, data blob)"""
    )
    cursor.execute(
        """CREATE UNIQUE INDEX IF NOT EXISTS idx_events_indices_id
            ON events_indices(chain_id,address,event,args)"""
    )

    # ERC20 metas table

    cursor.execute(
        """CREATE TABLE IF NOT EXISTS erc20_metas
                (chain_id integer, address text, name text, symbol text, decimals integer)"""
    )
    cursor.execute(
        """CREATE UNIQUE INDEX IF NOT EXISTS idx_erc20_metas_id
            ON erc20_metas(chain_id,address)"""
    )
    cursor.execute(
        """CREATE INDEX IF NOT EXISTS idx_erc20_metas_index
            ON erc20_metas(chain_id,address,symbol)"""
    )

    # Calls

    cursor.execute(
        """CREATE TABLE IF NOT EXISTS calls
                (chain_id integer, address text, calldata text, \
                block_number integer, response text)"""
    )
    cursor.execute(
        """CREATE UNIQUE INDEX IF NOT EXISTS idx_calls_id
            ON calls(chain_id,address,calldata,block_number)"""
    )

    # Balances table
    cursor.execute(
        """CREATE TABLE IF NOT EXISTS balances
                (chain_id integer, block_number integer, address text, balance integer)"""
    )
    cursor.execute(
        """CREATE UNIQUE INDEX IF NOT EXISTS idx_balances_id \
    ON balances(chain_id,block_number,address)
    """
    )

    conn.commit()



"""
Module for working with web3 datasets.

The datasets are fetched from blockchains via RPC endpoint, cached
and returned as a `polars.Dataframe <https://pola-rs.github.io/polars/py-polars/html/reference/dataframe.html>`_.

Note:
    Throughout the docs a notion of ``timepoint`` is used.
    ``timepoint`` is either block number or Unix timestamp or datetime.

Note:
    All values in the datasets are in natural units. For example ETHs,
    not WEIs.
"""

import polars as pl
from bokeh.io import curdoc

from web3cat.data.erc20s import ERC20Data
from web3cat.data.chainlink.chainlink_data import ChainlinkUSDData, ChainlinkData
from web3cat.data.ethers import EtherData
from web3cat.data.portfolios import PortfolioData
from web3cat.data.gearbox import GearboxData


pl.Config.set_fmt_str_lengths(44)
curdoc().theme = "dark_minimal"



from dataclasses import dataclass
from typing import List
from functools import cached_property
from datetime import datetime
from bokeh.plotting import Figure
from bokeh.models import GlyphRenderer
import numpy as np


from web3cat.fetcher.erc20_metas import ERC20Meta
from web3cat.data.chainlink import ChainlinkData
from web3cat.view.wireframes.abstract import TimeseriesWireframe


@dataclass(frozen=True)
class ChainlinkPricesWireframe(TimeseriesWireframe):
    """
    Chailink prices wireframe
    """

    token0: ERC20Meta
    token1: ERC20Meta

    @cached_property
    def data_key(self) -> str:
        return "chainlink_prices"

    @property
    def y_axis(self) -> str:
        return f"{self.token0.symbol.upper()} / {self.token1.symbol.upper()} price"

    def build_data(
        self, default_data: ChainlinkData | None, **core_args
    ) -> ChainlinkData:
        if default_data is None:
            return ChainlinkData(
                [self.token0.symbol, self.token1.symbol],
                self.start,
                self.end,
                **core_args,
            )
        default_data.add_token(self.token0.address)
        default_data.add_token(self.token1.address)
        return default_data

    def y(self, data: ChainlinkData) -> List[np.float64]:
        return data.prices(self.token0.address, self.token1.address, self.x(data))[
            "price"
        ].to_list()

    def plot(
        self, fig: Figure, x: List[datetime], y: List[np.float64], **kwargs
    ) -> GlyphRenderer:
        return fig.line(
            x,
            y,
            line_width=2,
            legend_label=self.y_axis,
            **kwargs,
        )



BLOCKS_PER_BIT = 1_000
"""
Defines the granularity of event indices. Blocks are divided into chunks.
Each chunk has the size of :const:`BLOCKS_PER_BIT`. If the bit is set,
the events for the whole chunk are fetched. Partial fetches are not
permitted.
"""



from typing import Any, Dict, Iterator, List, Tuple
from web3cat.fetcher.events.event import Event
from web3cat.fetcher.core import Core


class EventsRepo(Core):
    """
    Reading and writing :class:`Event` to database.
    """

    def find(
        self,
        event: str,
        address: str,
        argument_filters: Dict[str, Any] | None = None,
        from_block: int = 0,
        to_block: int = 2**32 - 1,
    ) -> Iterator[Event]:
        """
        Find all events in the database.

        Args:
            event: Event name
            address: Contract address
            argument_filters: an additional filter with keys being event fieds (AND query)
                              and values are filter values (list of values for OR query)
            from_block: starting from this block (inclusive)
            to_block: ending with this block (non-inclusive)

        Returns:
            Iterator over found events
        """
        cursor = self.conn.cursor()
        args_query, args_values = self._convert_filter_to_sql(argument_filters)
        statement = (
            "SELECT * FROM events WHERE chain_id = ? AND event = ? "
            f"AND address = ? AND block_number >= ? AND block_number < ?{args_query}"
        )
        args = tuple(
            [self.chain_id, event, address.lower(), from_block, to_block, *args_values]
        )
        cursor.execute(
            statement,
            args,
        )
        rows = cursor.fetchall()
        return (Event.from_row(r) for r in rows)

    def save(self, events: List[Event]):
        """
        Save a set of events into the database.

        Args:
            events: List of events to save
        """
        cursor = self.conn.cursor()
        rows = [e.to_row() for e in events]
        cursor.executemany(
            "INSERT INTO events VALUES(?,?,?,?,?,?,?) ON CONFLICT DO NOTHING", rows
        )

    def purge(self):
        """
        Clear all database entries
        """
        cursor = self.conn.cursor()
        cursor.execute("DELETE FROM events")

    def _convert_filter_to_sql(
        self, event_filter: Dict[str, Any] | None
    ) -> Tuple[str, List[Any]]:
        if event_filter is None:
            return ("", [])
        query = ""
        values = []
        for k, v in event_filter.items():
            if not isinstance(v, list):
                v = [v]
            inner = " OR ".join([f"""json_extract(args, "$.{k}") = ?""" for _ in v])
            query += f" AND ({inner})"
            values += v
        return (query, values)



from typing import List
from web3cat.fetcher.core import Core
from web3cat.fetcher.erc20_metas.erc20_meta import ERC20Meta


class ERC20MetasRepo(Core):
    """
    Reading and writing :class:`ERC20Meta` to database.
    """

    def find(self, token: str) -> ERC20Meta | None:
        """
        Find a :class:`ERC20Meta`.

        Args:
            token: token symbol or address

        Returns:
            An instance of :class:`ERC20Meta` or ``None`` if not found

        Examples:
            ::

                repo.find("DAI")
                repo.find("0x6B175474E89094C44Da98b954EedeAC495271d0F")
                # => same results
        """
        token = token.lower()
        cursor = self.conn.execute(
            "SELECT * FROM erc20_metas WHERE chain_id = ? AND (symbol = ? OR address = ?)",
            (self.chain_id, token, token),
        )
        row = cursor.fetchone()
        if not row:
            return None
        return ERC20Meta.from_row(row)

    def save(self, erc20_metas: List[ERC20Meta]):
        """
        Save :class:`ERC20Meta` to database.

        Args:
            erc20_metas: a list of :class:`ERC20Meta` to save
        """
        rows = [e.to_row() for e in erc20_metas]
        self.conn.executemany(
            "INSERT INTO erc20_metas VALUES(?,?,?,?,?) ON CONFLICT DO NOTHING", rows
        )

    def purge(self):
        """
        Clear all database entries
        """
        cursor = self.conn.cursor()
        cursor.execute("DELETE FROM erc20_metas")



from dataclasses import dataclass
from typing import List, Dict
from functools import cached_property
from datetime import datetime
from bokeh.plotting import Figure
from bokeh.models import GlyphRenderer
from bokeh.palettes import Category20
import numpy as np
from web3cat.fetcher.erc20_metas import ERC20Meta
from web3cat.fetcher.utils import short_address


from web3cat.data.portfolios import PortfolioData
from web3cat.view.wireframes.abstract import TimeseriesWireframe


@dataclass(frozen=True)
class PortfolioWireframe(TimeseriesWireframe):
    """
    Portfolio wireframe
    """

    tokens: List[ERC20Meta]
    base_token: ERC20Meta
    addresses: List[str]

    @cached_property
    def data_key(self) -> str:
        return "portfolio"

    @property
    def y_axis(self) -> str:
        return f"Balance ({self.base_token.symbol.upper()})"

    def x(self, data: PortfolioData) -> List[datetime]:
        return sorted(list(set(data.balances_and_prices["date"].to_list())))

    def build_data(
        self, default_data: PortfolioData | None, **core_args
    ) -> PortfolioData:
        if not default_data is None:
            raise ValueError("Only one porfolio view per figure is allowed")
        return PortfolioData(
            addresses=self.addresses,
            tokens=[token.symbol.upper() for token in self.tokens],
            base_tokens=[self.base_token.symbol.upper()],
            start=self.start,
            end=self.end,
            numpoints=self.numpoints,
            **core_args,
        )

    def y(self, data: PortfolioData) -> List[List[np.float64]]:
        raise ValueError("Unimplemented")

    def plot(
        self, fig: Figure, x: List[datetime], y: List[np.float64], **kwargs
    ) -> GlyphRenderer:
        raise ValueError("Unimplemented")


class PortfolioByAddressWireframe(PortfolioWireframe):
    """
    Portfolio by address
    """

    def y(self, data: PortfolioData) -> Dict[str, List[np.float64]]:
        series_dict = data.breakdown_by_address(
            self.base_token.symbol.upper()
        ).to_dict()
        exclude = ["timestamp", "date", "block_number"]
        return {k: v.to_list() for k, v in series_dict.items() if not k in exclude}

    def plot(
        self, fig: Figure, x: List[datetime], y: Dict[str, List[np.float64]], **kwargs
    ) -> GlyphRenderer:
        y = {**y}
        y.pop("total")

        colors = Category20[20] * (1 + len(y.keys()) // len(Category20[20]))
        colors = colors[: len(y.keys())]
        stackers = [*y.keys()]
        y["date"] = x
        args = {
            "stackers": stackers,
            "x": "date",
            "legend_label": [short_address(s) for s in stackers],
            "source": y,
            "alpha": 0.8,
            **kwargs,
            "color": colors,
        }
        return fig.varea_stack(**args)


class PortfolioByTokenWireframe(PortfolioWireframe):
    """
    Portfolio by token
    """

    def y(self, data: PortfolioData) -> Dict[str, List[np.float64]]:
        series_dict = data.breakdown_by_token(self.base_token.symbol.upper()).to_dict()
        exclude = ["timestamp", "date", "block_number"]
        return {k: v.to_list() for k, v in series_dict.items() if not k in exclude}

    def plot(
        self, fig: Figure, x: List[datetime], y: Dict[str, List[np.float64]], **kwargs
    ) -> GlyphRenderer:
        y = {**y}
        y.pop("total")

        colors = Category20[20] * (1 + len(y.keys()) // len(Category20[20]))
        colors = colors[: len(y.keys())]
        y["date"] = x
        stackers = [
            f"{meta.symbol.upper()} ({self.base_token.symbol.upper()})"
            for meta in self.tokens
        ]

        args = {
            "stackers": stackers,
            "x": "date",
            "legend_label": stackers,
            "source": y,
            "alpha": 0.8,
            **kwargs,
            "color": colors,
        }
        return fig.varea_stack(**args)


class PortfolioBalanceWireframe(PortfolioWireframe):
    """
    Portfolio token balances
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, tokens=[kwargs["base_token"]])

    def y(self, data: PortfolioData) -> Dict[str, List[np.float64]]:
        series_dict = data.balances(self.base_token.symbol.upper()).to_dict()
        exclude = ["timestamp", "date", "block_number"]
        return {k: v.to_list() for k, v in series_dict.items() if not k in exclude}

    def plot(
        self, fig: Figure, x: List[datetime], y: Dict[str, List[np.float64]], **kwargs
    ) -> GlyphRenderer:
        y = {**y}
        y.pop("total")

        colors = Category20[20] * (1 + len(y.keys()) // len(Category20[20]))
        colors = colors[: len(y.keys())]
        stackers = [*y.keys()]
        y["date"] = x
        args = {
            "stackers": stackers,
            "x": "date",
            "legend_label": [short_address(s) for s in stackers],
            "source": y,
            "alpha": 0.8,
            **kwargs,
            "color": colors,
        }
        return fig.varea_stack(**args)



from typing import Iterator, List
from web3cat.fetcher.blocks.block import Block
from web3cat.fetcher.core import Core


class BlocksRepo(Core):
    """
    Reading and writing :class:`Block` to database.
    """

    def get_block_after_timestamp(self, timestamp: int) -> Block | None:
        """
        Get the first block strictly after the timestamp in database.

        Args:
            chain_id: Ethereum chain_id
            timestamp: UNIX timestamp

        Returns:
            First block after the timestamp, ``None`` if the block doesn't exist
        """

        row = self.conn.execute(
            "SELECT * FROM blocks WHERE timestamp > ? AND chain_id = ? "
            "ORDER BY block_number LIMIT 1",
            (timestamp, self.chain_id),
        ).fetchone()
        if not row:
            return None
        return Block.from_row(row)

    def get_block_before_timestamp(self, timestamp: int) -> Block | None:
        """
        Get the first block at or before the timestamp in database.

        Args:
            chain_id: Ethereum chain_id
            timestamp: UNIX timestamp, UTC+0

        Returns:
            First block before the timestamp, ``None`` if the block doesn't exist
        """

        row = self.conn.execute(
            "SELECT * FROM blocks WHERE timestamp < ? AND chain_id = ? "
            "ORDER BY block_number DESC LIMIT 1",
            (timestamp, self.chain_id),
        ).fetchone()
        if not row:
            return None
        return Block.from_row(row)

    def find(self, blocks: int | List[int]) -> Iterator[Block]:
        """
        Find blocks by number

        Warning:
            The order of the blocks are not preserved. Duplicates are eliminated.

        Args:
            blocks: block number or a list of block numbers

        Returns:
            A list of found blocks
        """
        int_blocks = blocks if isinstance(blocks, list) else [blocks]
        int_blocks = list(set(int_blocks))
        if len(int_blocks) == 0:
            return iter([])

        statement = (
            "SELECT * FROM blocks WHERE block_number IN "
            f" ({','.join('?' * len(int_blocks))}) AND chain_id = ?"
        )
        out = self.conn.execute(
            statement,
            int_blocks + [self.chain_id],
        )
        out = (Block.from_row(b) for b in out)
        return sorted(out, key=lambda x: x.number)

    def all(self) -> Iterator[Block]:
        """
        All blocks in the database
        """
        blocks = self.conn.execute(
            "SELECT * FROM blocks WHERE chain_id = ?", (self.chain_id,)
        )
        return (Block.from_row(b) for b in blocks)

    def save(self, blocks: List[Block]):
        """
        Save a set of blocks into the database.

        Args:
            blocks: List of blocks to save
        """

        cursor = self.conn.cursor()
        rows = [b.to_row() for b in blocks]
        cursor.executemany(
            "INSERT INTO blocks VALUES(?,?,?) ON CONFLICT DO NOTHING", rows
        )

    def purge(self):
        """
        Clear all database entries
        """
        cursor = self.conn.cursor()
        cursor.execute("DELETE FROM blocks")



"""
Fetches ERC20 datasets and returns them as
`polars.Dataframe <https://pola-rs.github.io/polars/py-polars/html/reference/dataframe.html>`_.

Using :class:`ERC20Data` you can:

#. Get the timeseries of the wallet balances
#. Get all transfers for the wallet
#. Get top volumes breakdown by the wallet
#. etc.

Examples:

    .. code::

        from datetime import datetime
        from web3cat.data import ERC20Data

        dates = [datetime(2021, 6, 1), datetime(2021, 7, 1), datetime(2021, 8, 1), datetime(2021, 9, 1)]
        addresses = ["0x5d3a536E4D6DbD6114cc1Ead35777bAB948E3643", "0xbEbc44782C7dB0a1A60Cb6fe97d0b483032FF1C7"]
        erc20_data = ERC20Data("Dai", addresses, min(dates), max(dates))

        # All erc20 transfers for addresses
        erc20_data.transfers

        # All erc20 mints and burns
        erc20_data.emission

        # All token volumes by address
        erc20_data.volume

        # Historical total supply
        erc20_data.total_supply(dates)

        # Historical balances for addresses
        erc20_data.balances(addresses, dates)

"""

from web3cat.data.erc20s.erc20_data import ERC20Data



from hexbytes import HexBytes


class BitArray:
    """
    This is a simple bitset tailored for :mod:`events_indices` needs.

    Example:

        >>> bits = BitArray()
        >>> bits[3]
        False
        >>> bits[3] = True # 00010000
        >>> bits[0]
        False
        >>> bits[3]
        True
        >>> bits.prepend_empty_bytes(2) # 000000000000000000010000
        >>> bits[0]
        False
        >>> bits[3]
        False
        >>> bits[2 * 8 + 3]
        True
        >>> bits.set_range(0, 4, True) # 111100000000000000010000
        >>> bits[0]
        True
        >>> bits[1]
        True
        >>> bits[2]
        True
        >>> bits[3]
        True
        >>> bits[4]
        False

    """

    _data: bytearray

    def __init__(self, data: bytes | None = None):
        self._data = bytearray(data or [])

    def __getitem__(self, pos: int) -> bool:
        """
        Get the bit value at ``pos``

        Args:
            pos: The position of the bit

        Returns:
            Bit value (``False`` or ``True``)
        """
        byte_idx = pos // 8
        if byte_idx >= len(self._data):
            return False
        byte = self._data[pos // 8]
        return byte & (1 << (pos % 8)) != 0

    def __setitem__(self, pos: int, value: bool):
        """
        Set the bit value at ``pos``

        Args:
            pos: The position of the bit
            value: Bit value (``False`` or ``True``)
        """
        self._ensure_length(pos)
        if value:
            self._data[pos // 8] |= 1 << (pos % 8)
        else:
            self._data[pos // 8] &= 255 - (1 << (pos % 8))

    def __len__(self):
        return len(self._data) * 8

    def prepend_empty_bytes(self, num_bytes: int):
        """
        Adds zero bytes (8 * num_bytes ``False`` values) before the bitarray.

        Args:
            num_bytes: Number of bytes to add
        """
        new_data = bytearray([0] * num_bytes)
        for b in self._data:
            new_data.append(b)
        self._data = new_data

    def set_range(self, start: int, end: int, value: bool):
        """
        Set value for bits starting at ``start`` (inclusive) and ending
        at ``end`` (non-inclusive).

        Args:
            start: start of the range (inclusive)
            end: end of the range (non-inclusive)

        Exceptions:
            Both values should be >= 0 and ``end`` >= ``start``.
            Otherwise :class:`IndexError` is raised.

        """
        if start > end:
            raise IndexError(f"BitArray set_range with start: {start}, end: {end}")
        if end < 0:
            raise IndexError(f"BitArray set_range with start: {start}, end: {end}")
        self._ensure_length(end - 1)

        # inside the common byte or in neighbor bytes
        if end - start < 8:
            for i in range(start, end):
                self[i] = value
            return

        # at least in neighbor bytes
        full_start = start // 8 + 1
        full_end = end // 8
        val = 0 if value is False else 255
        for i in range(full_start, full_end):
            self._data[i] = val
        for i in range(start, full_start * 8):
            self[i] = value
        for i in range(full_end * 8, end):
            self[i] = value

    def to_hex(self) -> str:
        """
        Returns hex representation starting with '0x'.

        Example:

            >>> b = BitArray()
            >>> b.to_hex()
            '0x'
            >>> b[3] = True
            >>> b.to_hex()
            '0x08'
            >>> b[2] = True
            >>> b.to_hex()
            '0x0c'
            >>> b[4] = True
            >>> b.to_hex()
            '0x1c'
            >>> b[8] = True
            >>> b.to_hex()
            '0x1c01'

        """
        return HexBytes(self._data).hex()

    def _ensure_length(self, idx: int):
        while len(self._data) <= idx // 8:
            self._data.append(0)

    def __repr__(self):
        out = ""
        for b in self._data:
            for i in range(8):
                if b & (1 << i) > 0:
                    out += "1"
                else:
                    out += "0"
        return f"BitArray({out})"

    def __eq__(self, other):
        if type(other) is type(self):
            return self._data == other._data
        return False



from __future__ import annotations
from functools import cached_property
from typing import List
from datetime import datetime
from web3.constants import ADDRESS_ZERO
import polars as pl
from web3cat.fetcher.erc20_metas import ERC20Meta
from web3cat.data.ethers.ether_data import EtherData
from web3cat.data.erc20s import ERC20Data
from web3cat.data.chainlink import ChainlinkData
from web3cat.data.core import DataCore


class PortfolioData(DataCore):
    """
    Portfolio data (a group of tokens for a group of addresses)

    Arguments:
        tokens: A list of tokens to track
        base_tokens: A list of tokens for aggregation by price
        addresses: A list of addresses to track
        start: start of the data
        end: endof the data
        numpoints: number of points between ``start`` and ``end``
    """

    _tokens: List[str]
    _base_tokens: List[str]
    _addresses: List[str]
    _numpoints: int

    _erc20_datas: List[ERC20Data]
    _chainlink_data: ChainlinkData
    _ether_data: EtherData | None

    def __init__(
        self,
        tokens: List[str],
        base_tokens: List[str],
        addresses: List[str],
        start: int | datetime,
        end: int | datetime,
        numpoints: int,
        **kwargs,
    ):
        super().__init__(start, end, **kwargs)
        if len(tokens) == 0:
            raise ValueError("Tokens list cannot be empty")

        self._tokens = [token for token in tokens if token.upper() != "ETH"]
        if len(self._tokens) != tokens:
            self._tokens.append("ETH")
        self._base_tokens = base_tokens
        self._addresses = [addr.lower() for addr in addresses]
        self._numpoints = numpoints

        services = {
            "blocks_service": self._blocks_service,
            "balances_service": self._balances_service,
            "calls_service": self._calls_service,
            "erc20_metas_service": self._erc20_metas_service,
            "events_service": self._events_service,
        }

        self._erc20_datas = [
            ERC20Data(token, addresses, start, end, **services, **kwargs)
            for token in tokens
            if token.upper() != "ETH"
        ]
        self._chainlink_data = ChainlinkData(
            list(set(tokens + base_tokens)), start, end, **services, **kwargs
        )
        self._ether_data = None
        if "ETH" in [t.upper() for t in tokens]:
            self._ether_data = EtherData(start, end, **services, **kwargs)

    @cached_property
    def tokens(self) -> List[ERC20Meta]:
        """
        Tracked tokens metas.
        """
        return [self._get_meta(token) for token in self._tokens]

    @cached_property
    def base_tokens(self) -> List[ERC20Meta]:
        """
        Base tokens metas.
        """
        return [self._get_meta(token) for token in self._base_tokens]

    def balances(self, base_token: str) -> pl.DataFrame:
        """
        Breakdown of a specific token holdings by the owner address.

        Note:
            Unlike :meth:`breakdown_by_address` it doesn't convert
            other tokens. For example, with ``basetoken='USDC'`` ETH
            holdings will be just ignored.

        Arguments:
            base_token: token for holdings

        Returns:
            A Dataframe with fields:

        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | Field                | Type                       | Description                                                                  |
        +======================+============================+==============================================================================+
        | ``timestamp``        | :attr:`numpy.int64`        | Timestamp for the snapshot of the balance                                    |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``date``             | :class:`datetime.datetime` | Date for the timestamp                                                       |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``block_number``     | :class:`int`               | Number of the block                                                          |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``<address N>``      | :attr:`numpy.float64`      | Balance of the address of ``base_token``                                     |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``total``            | :attr:`numpy.float64`      | Total portfolio balance of ``base_token``                                    |
        +----------------------+----------------------------+------------------------------------------------------------------------------+

        """

        base_token = self._erc20_metas_service.get(base_token)
        dfs = self.balances_and_prices.partition_by("address", True)
        addr = dfs[0]["address"][0]
        out = dfs[0].rename({base_token.symbol.upper(): addr})[
            ["timestamp", "date", "block_number", addr]
        ]
        for df in dfs[1:]:
            addr = df["address"][0]
            df = df.rename({base_token.symbol.upper(): addr})[["block_number", addr]]
            out = out.join(df, on="block_number", how="left")
        out = out.with_column(pl.col(self._addresses[0]).alias("total"))
        for addr in self._addresses[1:]:
            out = out.with_column(pl.col("total") + pl.col(addr).alias("total"))
        return out

    def breakdown_by_address(self, base_token: str) -> pl.DataFrame:
        """
        Breakdown of token holdings by the owner address.

        Arguments:
            base_token: token for denomination of holdings

        Returns:
            A Dataframe with fields:

        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | Field                | Type                       | Description                                                                  |
        +======================+============================+==============================================================================+
        | ``timestamp``        | :attr:`numpy.int64`        | Timestamp for the snapshot of the balance                                    |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``date``             | :class:`datetime.datetime` | Date for the timestamp                                                       |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``block_number``     | :class:`int`               | Number of the block                                                          |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``<address N>``      | :attr:`numpy.float64`      | Balance of the address in ``base_token``                                     |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``total``            | :attr:`numpy.float64`      | Total portfolio balance in ``base_token``                                    |
        +----------------------+----------------------------+------------------------------------------------------------------------------+


        """
        base_token = self._erc20_metas_service.get(base_token)
        data = self._with_total_in(self.balances_and_prices, base_token)[
            [
                "timestamp",
                "date",
                "block_number",
                "address",
                f"total {base_token.symbol.upper()}",
            ]
        ]
        dfs = data.partition_by("address", True)
        addr = dfs[0]["address"][0]
        out = dfs[0].rename({f"total {base_token.symbol.upper()}": addr})[
            ["timestamp", "date", "block_number", addr]
        ]
        for df in dfs[1:]:
            addr = df["address"][0]
            df = df.rename({f"total {base_token.symbol.upper()}": addr})[
                ["block_number", addr]
            ]
            out = out.join(df, on="block_number", how="left")
        out = out.with_column(pl.col(self._addresses[0]).alias("total"))
        for addr in self._addresses[1:]:
            out = out.with_column(pl.col("total") + pl.col(addr).alias("total"))
        return out

    def breakdown_by_token(self, base_token: str) -> pl.DataFrame:
        """
        Breakdown of token holdings by the owner address.

        Arguments:
            base_token: token for denomination of holdings

        Returns:
            A Dataframe with fields:

        +--------------------------------+----------------------------+------------------------------------------------------------------------------+
        | Field                          | Type                       | Description                                                                  |
        +================================+============================+==============================================================================+
        | ``timestamp``                  | :attr:`numpy.int64`        | Timestamp for the snapshot of the balance                                    |
        +--------------------------------+----------------------------+------------------------------------------------------------------------------+
        | ``date``                       | :class:`datetime.datetime` | Date for the timestamp                                                       |
        +--------------------------------+----------------------------+------------------------------------------------------------------------------+
        | ``block_number``               | :class:`int`               | Number of the block                                                          |
        +--------------------------------+----------------------------+------------------------------------------------------------------------------+
        | ``<token> (<base_token>)``     | :attr:`numpy.float64`      | Value of token in ``base_token``                                             |
        +--------------------------------+----------------------------+------------------------------------------------------------------------------+
        | ``total``                      | :attr:`numpy.float64`      | Total portfolio balance in ``base_token``                                    |
        +--------------------------------+----------------------------+------------------------------------------------------------------------------+
        """
        base_token = self._erc20_metas_service.get(base_token)
        data = self.balances_and_prices.with_column(
            (
                pl.col(self.tokens[0].symbol.upper())
                * pl.col(
                    f"{self.tokens[0].symbol.upper()} / {base_token.symbol.upper()}"
                )
            ).alias(f"{self.tokens[0].symbol.upper()} ({base_token.symbol.upper()})")
        )
        for token in self.tokens[1:]:
            data = data.with_column(
                (
                    pl.col(token.symbol.upper())
                    * pl.col(f"{token.symbol.upper()} / {base_token.symbol.upper()}")
                ).alias(f"{token.symbol.upper()} ({base_token.symbol.upper()})")
            )
        column_names = ["timestamp", "date", "block_number"]
        agg_column_names = [
            pl.col(f"{t.symbol.upper()} ({base_token.symbol.upper()})").sum()
            for t in self.tokens
        ]
        column_names += agg_column_names
        data = (
            data.groupby(["timestamp", "date", "block_number"])
            .agg(agg_column_names)
            .sort(["timestamp"])
        )
        data = data.with_column(
            pl.col(
                f"{self.tokens[0].symbol.upper()} ({base_token.symbol.upper()})"
            ).alias("total")
        )
        for t in self.tokens[1:]:
            data = data.with_column(
                (
                    pl.col("total")
                    + pl.col(f"{t.symbol.upper()} ({base_token.symbol.upper()})")
                ).alias("total")
            )
        return data

    @cached_property
    def balances_and_prices(self) -> pl.DataFrame:
        """
        Balances and prices data.

        +--------------------------------+----------------------------+------------------------------------------------------------------------------+
        | Field                          | Type                       | Description                                                                  |
        +================================+============================+==============================================================================+
        | ``timestamp``                  | :attr:`numpy.int64`        | Timestamp for the snapshot of the balance                                    |
        +--------------------------------+----------------------------+------------------------------------------------------------------------------+
        | ``date``                       | :class:`datetime.datetime` | Date for the timestamp                                                       |
        +--------------------------------+----------------------------+------------------------------------------------------------------------------+
        | ``block_number``               | :class:`int`               | Number of the block                                                          |
        +--------------------------------+----------------------------+------------------------------------------------------------------------------+
        | ``address``                    | :class:`str`               | Address of tokens owner                                                      |
        +--------------------------------+----------------------------+------------------------------------------------------------------------------+
        | ``<token>``                    | :attr:`numpy.float64`      | Value of token in ``base_token``                                             |
        +--------------------------------+----------------------------+------------------------------------------------------------------------------+
        | ``<token> / <base_token>``     | :attr:`numpy.float64`      | Price of token in terms of base token                                        |
        +--------------------------------+----------------------------+------------------------------------------------------------------------------+
        | ``total``                      | :attr:`numpy.float64`      | Total portfolio balance in ``base_token``                                    |
        +--------------------------------+----------------------------+------------------------------------------------------------------------------+

        """
        step = int(
            (self.to_block_number - self.from_block_number) / (self._numpoints - 1)
        )
        block_numbers = list(range(self.from_block_number, self.to_block_number, step))
        if block_numbers[-1] != self.to_block_number:
            block_numbers.append(self.to_block_number)

        data = (
            self._erc20_datas[0]
            .balances(self._addresses, block_numbers)
            .rename({"balance": self._tokens[0]})
        )
        for i in range(1, len(self._erc20_datas)):
            erc20 = self._erc20_datas[i]
            balances = erc20.balances(self._addresses, block_numbers)[
                ["block_number", "address", "balance"]
            ].rename({"balance": self._tokens[i]})
            data = data.join(balances, on=["block_number", "address"], how="left")

        if not self._ether_data is None:
            balances = self._ether_data.balances(self._addresses, block_numbers)[
                ["block_number", "address", "balance"]
            ].rename({"balance": "ETH"})
            data = data.join(balances, on=["block_number", "address"], how="left")

        for token in self.tokens:
            for base_token in self.base_tokens:
                prices = self._chainlink_data.prices(
                    token.symbol.upper(), base_token.symbol.upper(), block_numbers
                )
                prices = prices[["block_number", "price"]].rename(
                    {"price": f"{token.symbol.upper()} / {base_token.symbol.upper()}"}
                )
                data = data.join(prices, on=["block_number"], how="left")

        return data.sort("block_number")

    def _with_total_in(self, data: pl.DataFrame, base_token: ERC20Meta) -> pl.DataFrame:
        name = f"total {base_token.symbol.upper()}"
        data = data.with_column(
            (
                pl.col(self.tokens[0].symbol.upper())
                * pl.col(
                    f"{self.tokens[0].symbol.upper()} / {base_token.symbol.upper()}"
                )
            ).alias(name)
        )
        for token in self.tokens[1:]:
            data = data.with_column(
                (
                    pl.col(name)
                    + pl.col(token.symbol.upper())
                    * pl.col(f"{token.symbol.upper()} / {base_token.symbol.upper()}")
                ).alias(name)
            )
        return data

    def _get_meta(self, token: str) -> ERC20Meta | None:
        if token.upper() == "USD":
            return ERC20Meta(1, ADDRESS_ZERO, "USD", "USD", 6, None)
        return self._erc20_metas_service.get(token)



"""
Module for fetching and caching contract calls.

The main class of this module is :class:`CallsService`.
It is used for making static calls to Ethereum contracts
and caching them for subsequent calls.

Example:
    ::

        from web3cat.fetcher.calls import CallsService
        from web3cat.fetcher.erc20_metas import ERC20MetasService

        erc20 = ERC20MetasService.create()
        dai = erc20.get("Dai")
        compound_address = "0x5d3a536E4D6DbD6114cc1Ead35777bAB948E3643"

        service = CallsService.create()
        response = service.get_call(
            dai.contract.functions.balanceOf(compound_address), block_number=15632000
        )
        # => going for web3 rpc

        response = service.get_call(
            dai.contract.functions.balanceOf(compound_address), block_number=15632000
        )
        # => serving from cache
# """


from web3cat.fetcher.calls.call import Call
from web3cat.fetcher.calls.repo import CallsRepo
from web3cat.fetcher.calls.service import CallsService



"""
Utility functions.
"""

import sys
from typing import Any, Dict, Union
import json
from hexbytes import HexBytes
from eth_typing.encoding import HexStr
from web3.datastructures import AttributeDict
from web3.contract import ContractFunction, get_abi_input_types
from web3.auto import w3
from eth_utils import function_abi_to_4byte_selector
import os


class Web3JsonEncoder(json.JSONEncoder):
    """
    Custom encoder to parse `Web3 <https://web3py.readthedocs.io/en/stable/>`_ responses.
    By default `Web3 <https://web3py.readthedocs.io/en/stable/>`_ returns
    responses as ``AttributeDict`` with binary values.
    :class:`Web3JsonEncoder` thansforms it into the :code:`0x...`
    hex format.
    """

    def default(self, o: Any) -> Union[Dict[Any, Any], HexStr]:
        """
        Convert Web3 response to ``dict``
        """
        if isinstance(o, AttributeDict):
            return dict(o.items())
        if isinstance(o, HexBytes):
            return HexStr(o.hex())
        if isinstance(o, (bytes, bytearray)):
            return HexStr(HexBytes(o).hex())
        return json.JSONEncoder.default(self, o)


def json_response(response: AttributeDict) -> str:
    """
    Convert ``AttributeDict`` to standard json string

    Args:
        response: a `Web3 <https://web3py.readthedocs.io/en/stable/>`_ response

    Returns:
        json string
    """
    return json.dumps(response, cls=Web3JsonEncoder)


def short_address(address: str) -> str:
    """
    Converts ethereum address to short version (for display purposes only).

    Args:
        address: Ethereum address to shorten

    Returns:
        Short version of the address.

    Examples:
        ::

            print(short_address("0x6B175474E89094C44Da98b954EedeAC495271d0F"))
            # 0x6B17...1d0F

    """
    return f"{address[:6]}...{address[38:]}"


def calldata(call: ContractFunction) -> str:
    """
    Hex calldata for :class:`web3.contract.ContractFunction` call

    Args:
        call: A web3 call

    Returns:
        Hex data (starting with 0x, lowercase)
    """

    selector = HexBytes(function_abi_to_4byte_selector(call.abi)).hex()
    abi_types = get_abi_input_types(call.abi)
    bytes_calldata = w3.codec.encode(abi_types, call.args)
    return selector + HexBytes(bytes_calldata).hex()[2:]


LAST_PROGRESS_BAR_LENGTH = 0
LAST_PERCENT = 0


def print_progress(
    iteration: int,
    total: int,
    prefix: str = "",
    suffix: str = "",
    decimals: int = 1,
    bar_length: int = 20,
) -> int:
    """
    Call in a loop to create terminal progress bar.

    Args:
        iteration: current iteration
        total: total iterations
        prefix: prefix string
        suffix: suffix string
        decimals: positive number of decimals in percent complete
        bar_length: character length of bar
    """
    enabled = os.environ.get("WEB3_DISABLE_STDOUT")
    if not enabled is None:
        return

    global LAST_PROGRESS_BAR_LENGTH
    global LAST_PERCENT

    # Avoid too much output
    if (
        iteration != total
        and iteration > 0
        and iteration / total - LAST_PERCENT < 0.001
    ):
        return

    LAST_PERCENT = iteration / total
    str_format = "{0:." + str(decimals) + "f}"
    percents = str_format.format(100 * (iteration / float(total)))
    filled_length = int(round(bar_length * iteration / float(total)))
    vert_bar = "█" * filled_length + "-" * (bar_length - filled_length)
    text = f"{prefix} |{vert_bar}| {percents}% {suffix}\r"
    _ = sys.stdout.write(f"{' ' * LAST_PROGRESS_BAR_LENGTH}\r")
    _ = sys.stdout.write(text)
    LAST_PROGRESS_BAR_LENGTH = len(text)

    if iteration == total:
        sys.stdout.write("\n")
        sys.stdout.flush()
        LAST_PERCENT = 0
        LAST_PROGRESS_BAR_LENGTH = 0



from typing import Any, Dict, Iterator, List

import json
from web3cat.fetcher.events_indices.index import EventsIndex
from web3cat.fetcher.core import Core


class EventsIndicesRepo(Core):
    """
    Reading and writing :class:`EventsIndex` to database.
    """

    def find_indices(
        self,
        address: str,
        event: str,
        args: Dict[str, Any] | None = None,
    ) -> Iterator[EventsIndex]:
        """
        Find all indices that match :code:`chain_id`, ``address``,
        ``event``, and ``args``.

        The match for ``args`` is non-trivial.
        It tries to find all indices that could have the events for
        the specific ``args``.

        Example
        ~~~~~~~

        Imagine two indices for the ERC20 Transfer event:

        1. Stating that all ``Transfer`` events were fetched from block 2000 to 4000
        2. Stating that ``Transfer`` from address "0x6b17..." events were fetched
           from block 3000 to 4000

        Now we want to query if the ``Transfer`` events were fetched for
        :code:`{"from": "0x6b17..."}` for blocks 2500 to 4000. A naive implementation
        would return just the :code:`{"from": "0x6b17..."}` index stating that blocks
        2500 to 3000 are missing. However, these events were already fetched
        as part of all ``Transfer`` events from block 2000 to 4000.
        That's why a list of indices is returned, so the check is made
        against the data that is really missing.

        Args:
            address: Contract address
            event: Event name
            args: Argument filters

        Returns:
            Iterator of matched indices
        """
        rows = self.conn.execute(
            "SELECT * FROM events_indices WHERE chain_id = ? "
            "AND address = ? AND event = ?",
            (self.chain_id, address, event),
        )
        indices = (EventsIndex.from_row(r) for r in rows)
        return (i for i in indices if is_softer_filter_than(i.args, args))

    def get_index(
        self,
        address: str,
        event: str,
        args: Dict[str, Any] | None = None,
    ) -> EventsIndex | None:
        """
        Find an index with the exact match of :code:`chain_id`, ``address``,
        ``event``, and ``args``.

        Args:
            chain_id: Ethereum chain_id
            address: Contract address
            event: Event name
            args: Argument filters

        Returns:
            Found index, or ``None`` if nothing is found
        """

        args = args or {}
        cursor = self.conn.cursor()
        cursor.execute(
            "SELECT * FROM events_indices WHERE chain_id = ? AND address = ? "
            "AND event = ? and args = ?",
            (
                self.chain_id,
                address,
                event,
                json.dumps(EventsIndex.normalize_args(args)),
            ),
        )
        row = cursor.fetchone()
        if row is None:
            return None
        return EventsIndex.from_row(row)

    def save(self, indices: List[EventsIndex]):
        """
        Save indices to database.

        Args:
            indices: a list of indices to save
        """
        cursor = self.conn.cursor()
        rows = [i.to_row() for i in indices]
        statement = (
            "INSERT INTO events_indices VALUES (?,?,?,?,?) "
            "ON CONFLICT(chain_id,address,event,args) DO UPDATE SET data = excluded.data"
        )
        cursor.executemany(
            statement,
            rows,
        )

    def purge(self):
        """
        Clear all database entries
        """
        cursor = self.conn.cursor()
        cursor.execute("DELETE FROM events_indices")


def is_softer_filter_than(filter1: Any | None, filter2: Any | None) -> bool:
    """
    Checks if ``filter1`` is arguments filter is a softer version of
    ``filter2`` argument filter. Softer means more results after
    filtering.

    Args:
        filter1: Argument filters to check. The convention is :code:`None == {}`
        filter2: Argument filters to check. The convention is :code:`None == {}`
    """
    if filter1 is None:
        return True
    if filter2 is None:
        if isinstance(filter1, dict) and len(filter1.keys()) == 0:
            return True
        return False
    if isinstance(filter1, dict):
        if not isinstance(filter2, dict):
            return False
        for key in filter1.keys():
            if not key in filter2:
                return False
            if not is_softer_filter_than(filter1[key], filter2[key]):
                return False
        return True
    if isinstance(filter1, list):
        if not isinstance(filter2, list):
            return False
        sb = set(filter1)
        sp = set(filter2)
        return sp.issubset(sb)
    return filter1 == filter2



"""
Chainlink prices datasets.

The module has two classes:

* :class:`ChainlinkData` is a price data for a pair of tokens
* :class:`ChainlinkUSDData` is a price data of a token for usd

:class:`ChainlinkUSDData` is a strictly less powerful data source.
However, it's also more lightweight than :class:`ChainlinkData`.
The latter has two :class:`ChainlinkUSDData` under the hood and
relative price tokens are determined using cross exchange rates
to USD.

Note:
    Expect the error of up to 2% in this price data (this is
    how chainlink oracles work)

Note:
    Current oracles data are generally available from mid-2021.

Examples:

    .. code::

        from datetime import datetime
        from web3cat.data import ChainlinkData

        dates = [datetime(2021, 6, 1), datetime(2021, 7, 1), datetime(2021, 8, 1), datetime(2021, 9, 1)]
        tokens = ["USDC", "WETH", "WBTC"]
        chainlink_data = ChainlinkData(tokens, min(dates), max(dates))

        # Historical prices WETH/USDC
        chainlink_data.prices("WETH", "USDC", dates)

        # Historical prices WBTC/WETH
        chainlink_data.prices("WBTC", "WETH", dates)

        # Underlying USDC/USD prices
        chainlink_data.get_data("USDC").prices(dates)
"""

from web3cat.data.chainlink.chainlink_data import ChainlinkUSDData, ChainlinkData



from __future__ import annotations
from datetime import datetime
from functools import cached_property
from typing import Any, Dict, List
from bokeh.plotting import Figure, show
from bokeh.models import (
    BasicTickFormatter,
    NumeralTickFormatter,
    Range1d,
    LinearAxis,
)
from bokeh.plotting import figure
from bokeh.palettes import Category10
import numpy as np
from web3.constants import ADDRESS_ZERO

from web3cat.fetcher.erc20_metas import ERC20MetasService, ERC20Meta
from web3cat.view.wireframes import (
    Wireframe,
    BalanceWireframe,
    TotalSupplyWireframe,
    ChainlinkPricesWireframe,
    EthBalanceWireframe,
    PortfolioByAddressWireframe,
    PortfolioByTokenWireframe,
    PortfolioBalanceWireframe,
)


class View:
    """
    Visualization of the blockchain data

    Arguments:
        **defaults: default values for all methods called
        **common_defaults: Common defaults for all graphs.
                           Currently it's ``x_axis_name`` and ``y_axis_name``.
        **fig_args: the arguments for the :class:`bokeh.plotting.Figure`
        **core_args: the arguments for the :class:`fetcher.core.Core`
    """

    _wireframes: List[Wireframe]
    _core_args: Dict[str, Any]
    _fig_args: Dict[str, Any]
    _defaults: Dict[str, Any]
    _colors: List[Any]

    _erc20_metas_service: ERC20MetasService
    _glyphs: List[Any]
    _datas: Dict[str, Any]

    def __init__(self, **kwargs):
        self._wireframes = []
        self._datas = {}
        self._colors = kwargs.pop("colors", Category10[10])
        self._glyphs = []
        self._core_args = {
            k: kwargs.pop(k)
            for k in ["rpc", "cache_path", "block_grid_step", "w3", "conn"]
            if k in kwargs
        }
        self._fig_args = {
            k: kwargs.pop(k)
            for k in [
                "x_range",
                "y_range",
                "x_axis_type",
                "y_axis_type",
                "tools",
                "x_minor_ticks",
                "y_minor_ticks",
                "x_axis_location",
                "y_axis_location",
                "x_axis_label",
                "y_axis_label",
                "active_drag",
                "active_inspect",
                "active_scroll",
                "active_tap",
                "active_multi",
                "tooltips",
            ]
            if k in kwargs
        }
        self._erc20_metas_service = ERC20MetasService.create(**self._core_args)
        self._defaults = {"numpoints": 100, **kwargs}

    def get_data(self, index: int) -> Any:
        """
        Get the underlying data.
        """
        return self._datas[self._wireframes[index].data_key]

    @cached_property
    def figure(self) -> Figure:
        """
        :class:`bokeh.plotting.Figure` for the view.
        """
        return figure(
            toolbar_location="above",
            tools="pan,wheel_zoom,hover,reset,save",
            x_axis_type=self._wireframes[0].x_axis_name or self._wireframes[0].x_axis,
            height=400,
            width=int(400 * 2.3),
            **self._fig_args,
        )

    def total_supply(
        self,
        token: str | None = None,
        start: int | datetime | None = None,
        end: int | datetime | None = None,
        numpoints: int | None = None,
        **kwargs,
    ) -> View:
        """
        Historical total supply of the ERC20 tokens

        Args:
            token: ERC20 token
            start: start timepoint
            end: end timepoint
            numpoints: number of points in between
            kwargs: common_defaults, see :class:`View`
        """
        args = self._build_wireframe_args(
            {
                "token": token,
                "start": start,
                "end": end,
                "numpoints": numpoints,
                **kwargs,
            }
        )
        args["token"] = self._erc20_metas_service.get(args["token"])
        self._wireframes.append(TotalSupplyWireframe(**args))
        return self

    def chainlink_prices(
        self,
        token: str | None = None,
        base_token: str | None = None,
        start: int | datetime | None = None,
        end: int | datetime | None = None,
        numpoints: int | None = None,
        **kwargs,
    ):
        """
        Chainlink prices.

        Args:
            token: Numerator in the price
            base_token: Denominator in the price
            start: start timepoint
            end: end timepoint
            numpoints: number of points in between
            kwargs: common_defaults, see :class:`View`
        """
        args = self._build_wireframe_args(
            {
                "token": token,
                "base_token": base_token,
                "start": start,
                "end": end,
                "numpoints": numpoints,
                **kwargs,
            }
        )
        for name in ["token", "base_token"]:
            if args[name].upper() == "USD":
                args[name] = ERC20Meta(1, ADDRESS_ZERO, "USD", "USD", 6, None)
            else:
                args[name] = self._erc20_metas_service.get(args[name])
        args["token0"] = args.pop("token")
        args["token1"] = args.pop("base_token")
        self._wireframes.append(ChainlinkPricesWireframe(**args))
        return self

    def balance(
        self,
        address: str | List[str] | None = None,
        token: str | None = None,
        start: int | datetime | None = None,
        end: int | datetime | None = None,
        numpoints: int | None = None,
        **kwargs,
    ):
        """
        Balances of ERC20 token.

        Args:
            address: an address or the list of addresses
            token: ERC20 token
            start: start timepoint
            end: end timepoint
            numpoints: number of points in between
            kwargs: common_defaults, see :class:`View`
        """
        args = self._build_wireframe_args(
            {
                "token": token,
                "address": address,
                "start": start,
                "end": end,
                "numpoints": numpoints,
                **kwargs,
            }
        )
        if isinstance(args["address"], str):
            args["address"] = [args["address"]]
        for addr in args["address"]:
            args_item = {**args, "address": addr}
            if args_item["token"].upper() == "ETH":
                args_item.pop("token")
                self._wireframes.append(EthBalanceWireframe(**args_item))
            else:
                args_item["token"] = self._erc20_metas_service.get(args_item["token"])
                self._wireframes.append(BalanceWireframe(**args_item))
        return self

    def portfolio_by_address(
        self,
        tokens: List[str] | None = None,
        base_token: str | None = None,
        addresses: List[str] | None = None,
        start: int | datetime | None = None,
        end: int | datetime | None = None,
        numpoints: int | None = None,
        **kwargs,
    ) -> View:
        """
        Portfolio of ERC20 tokens, breakdown by address.

        Note:
            It can be called only once per the :class:`View`

        Args:
            tokens: a list of token holding that needs to be analyzed
            base_token: all values will be aggregated in this token
            start: start timepoint
            end: end timepoint
            numpoints: number of points in between
            kwargs: common_defaults, see :class:`View`
        """

        args = self._build_wireframe_args(
            {
                "tokens": tokens,
                "base_token": base_token,
                "addresses": addresses,
                "start": start,
                "end": end,
                "numpoints": numpoints,
                **kwargs,
            }
        )
        token_metas = [self._erc20_metas_service.get(token) for token in args["tokens"]]
        base_token_meta = self._erc20_metas_service.get(args["base_token"])
        addresses = [a.lower() for a in (args["addresses"] or [])]
        args["tokens"] = token_metas
        args["base_token"] = base_token_meta
        args["addresses"] = addresses

        self._wireframes.append(PortfolioByAddressWireframe(**args))

        return self

    def portfolio_by_token(
        self,
        tokens: List[str] | None = None,
        base_token: str | None = None,
        addresses: List[str] | None = None,
        start: int | datetime | None = None,
        end: int | datetime | None = None,
        numpoints: int | None = None,
        **kwargs,
    ):
        """
        Portfolio of ERC20 tokens, breakdown by token.

        Note:
            It can be called only once per the :class:`View`

        Args:
            tokens: a list of token holding that needs to be analyzed
            base_token: all values will be aggregated in this token
            start: start timepoint
            end: end timepoint
            numpoints: number of points in between
            kwargs: common_defaults, see :class:`View`
        """

        args = self._build_wireframe_args(
            {
                "tokens": tokens,
                "base_token": base_token,
                "addresses": addresses,
                "start": start,
                "end": end,
                "numpoints": numpoints,
                **kwargs,
            }
        )
        token_metas = [self._erc20_metas_service.get(token) for token in args["tokens"]]
        base_token_meta = self._erc20_metas_service.get(args["base_token"])
        addresses = [a.lower() for a in (args["addresses"] or [])]
        args["tokens"] = token_metas
        args["base_token"] = base_token_meta
        args["addresses"] = addresses

        self._wireframes.append(PortfolioByTokenWireframe(**args))

        return self

    def portfolio_balance(
        self,
        base_token: str | None = None,
        addresses: List[str] | None = None,
        start: int | datetime | None = None,
        end: int | datetime | None = None,
        numpoints: int | None = None,
        **kwargs,
    ) -> View:
        """
        Portfolio of one specific ERC20 token, breakdown by address.

        Note:
            It can be called only once per the :class:`View`

        Args:
            base_token: a token tracked for balances
            start: start timepoint
            end: end timepoint
            numpoints: number of points in between
            kwargs: common_defaults, see :class:`View`
        """

        args = self._build_wireframe_args(
            {
                "base_token": base_token,
                "addresses": addresses,
                "start": start,
                "end": end,
                "numpoints": numpoints,
                **kwargs,
            }
        )
        base_token_meta = self._erc20_metas_service.get(args["base_token"])
        addresses = [a.lower() for a in (args["addresses"] or [])]
        args["base_token"] = base_token_meta
        args["addresses"] = addresses

        self._wireframes.append(PortfolioBalanceWireframe(**args))

        return self

    def show(self):
        """
        Fetch the data and output the figure.
        """
        self._build_data()
        for wf in self._wireframes:
            data = self._datas[wf.data_key]
            x = wf.x(data)
            y = wf.y(data)
            if isinstance(y, dict):
                for y_sub in y.values():
                    self._update_axes(self.figure, min(y_sub), max(y_sub), wf)
            else:
                if len(y) > 0:
                    if isinstance(y[0], list):
                        for y_sub in y:
                            self._update_axes(self.figure, min(y_sub), max(y_sub), wf)
                    else:
                        self._update_axes(self.figure, min(y), max(y), wf)

            self._glyphs.append(
                wf.plot(
                    self.figure,
                    x,
                    y,
                    color=self._get_color(),
                    y_range_name=wf.y_axis_name or wf.y_axis,
                )
            )
        show(self.figure)

    def _build_data(self):
        for wf in self._wireframes:
            data = self._datas.get(wf.data_key, None)
            self._datas[wf.data_key] = wf.build_data(data, **self._core_args)

    def _build_wireframe_args(self, args: Dict[str, Any]):
        args_filter = {*args.keys(), "x_axis_name", "y_axis_name"}
        clean_args = {k: v for k, v in args.items() if not v is None}
        merged = {**self._defaults, **clean_args}
        return {k: v for k, v in merged.items() if k in args_filter}

    def _get_color(self):
        color = self._colors[len(self._glyphs) % len(self._colors)]
        return color

    def _update_axes(
        self, fig: Figure, miny: float, maxy: float, wf: Wireframe
    ) -> Dict[str, Any]:
        miny /= 1.05
        maxy *= 1.05
        y_axis = wf.y_axis_name or wf.y_axis
        formatter = BasicTickFormatter()
        if (miny + maxy) / 2 <= 0:
            order = 0
        else:
            order = int(np.log((miny + maxy) / 2) / np.log(10))
        if order <= 0:
            zeroes = -order + 2
        if order > 0:
            zeroes = 3
        if order > 3:
            zeroes = 2
        if order > 6:
            zeroes = 3
        fmt = f"0.{''.join(['0'] * zeroes)}a"
        formatter = NumeralTickFormatter(format=fmt)

        fig.extra_y_ranges = fig.extra_y_ranges or {}
        needs_y_axis_init = len(fig.extra_y_ranges.keys()) == 0
        needs_y_axis_add = not y_axis in fig.extra_y_ranges
        old_range = fig.extra_y_ranges.get(y_axis, Range1d(miny, maxy))
        new_range = Range1d(min(miny, old_range.start), max(maxy, old_range.end))
        fig.extra_y_ranges[y_axis] = new_range

        if needs_y_axis_init:
            fig.yaxis[0].y_range_name = y_axis
            fig.yaxis[0].axis_label = y_axis
            fig.yaxis[0].formatter = formatter
        elif needs_y_axis_add:
            loc = "left" if len(fig.yaxis) % 2 == 0 else "right"
            fig.add_layout(
                LinearAxis(
                    y_range_name=y_axis,
                    axis_label=y_axis,
                    formatter=formatter,
                ),
                loc,
            )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['web3cat',
 'web3cat.data',
 'web3cat.data.chainlink',
 'web3cat.data.erc20s',
 'web3cat.data.ethers',
 'web3cat.data.gearbox',
 'web3cat.data.portfolios',
 'web3cat.fetcher',
 'web3cat.fetcher.balances',
 'web3cat.fetcher.blocks',
 'web3cat.fetcher.calls',
 'web3cat.fetcher.erc20_metas',
 'web3cat.fetcher.events',
 'web3cat.fetcher.events_indices',
 'web3cat.view',
 'web3cat.view.wireframes']

package_data = \
{'': ['*'], 'web3cat.data.gearbox': ['abi/*']}

install_requires = \
['bokeh>=2.4.3,<3.0.0',
 'jupyter-bokeh>=3.0.5,<4.0.0',
 'numpy>=1.23.3,<2.0.0',
 'polars>=0.14.18,<0.15.0',
 'web3>=6.0.0-beta.8,<7.0.0']

setup_kwargs = {
    'name': 'web3cat',
    'version': '0.1.5',
    'description': 'Data visualization for EVM-based blockchains',
    'long_description': '.. image:: https://readthedocs.org/projects/web3cat/badge/?version=latest\n    :target: https://web3cat.readthedocs.io/en/latest/?badge=latest\n    :alt: Documentation Status\n\nWelcome to Web3cat!\n===================\n\n.. image:: /docs/images/web3cat.png\n    :width: 180\n\n**Meow, fellow researcher!**\n\nWeb3cat is a framework for fetching and analyzing blockchain data.\n\nCurrently, it supports only EVM chains: `Ethereum <https://ethereum.org>`_,\n`Polygon <https://polygon.technology>`_, `BNB <https://www.bnbchain.org>`_, etc.\n\nDesign philosophy\n-----------------\n\n#. Visualize the data with minimum code\n#. Free for downloading, saving, and analyzing the data however you want\n#. Cache the data to avoid re-fetching anything at all\n#. Fully decentralized, that is, depending only on the blockchain RPC\n\n\nQuick demo\n----------\n.. image:: /docs/images/web3cat_demo.gif\n\n\nGetting started\n---------------\n\n1. Install python package \n\n.. code::\n\n    pip install web3cat\n\n2. Set up your archive node rpc. The easiest (and also free) way is to use\n   `Alchemy <https://alchemy.com>`_.\n\n3. Set initial configuration\n\n.. code::\n\n    import os\n    os.environ[\'WEB3_PROVIDER_URI\'] = \'https://eth-mainnet.g.alchemy.com/v2/<YOUR_ALCHEMY_API_KEY>\'\n    os.environ[\'WEB3_CACHE_PATH\']="cache.sqlite3"\n\n4. (optional for Jupyter) Initialize bokeh for python notebooks\n\n.. code::\n\n    from bokeh.io import output_notebook\n\n    output_notebook()\n\n5. Run sample visualization\n\n.. code::\n\n    from web3cat.view import View\n    from datetime import datetime\n\n    v = View(token="DAI", start=datetime(2022, 6, 1), end = datetime(2022, 10, 30)) \\\n        .total_supply() \\\n        .balance(["0x5d3a536E4D6DbD6114cc1Ead35777bAB948E3643", "0xbEbc44782C7dB0a1A60Cb6fe97d0b483032FF1C7"])\n    v.show()\n\n.. image:: docs/images/view1.png\n\n6. Get underlying data\n\n    .. code::\n\n        v.get_data(0).transfers[["date", "block_number", "from", "to", "value"]]\n\n    .. image:: docs/images/view_getting_started1.png\n\nRoadmap\n-------\n\nUse framework base layers to add analytics for protocols like\nUniswap, Liquity, Aave, Compound, Frax, etc.\n\nContributing\n------------\n\nSo far no bureaucracy here, open issues, make pull requests, and have fun!',
    'author': 'Mono6',
    'author_email': 'monowhile6@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
